package spam.evaluation;

import spam.exception.PatternTreeFactoryException;
import spam.exception.UnsupportedOperatorForPatternTreeException;
import spam.model.TreePattern;
import spam.model.TreePatternFactory;
import spam.model.TreePatternNode;
import spam.service.WellDesignedService;

import com.hp.hpl.jena.sparql.algebra.Op;
import com.hp.hpl.jena.sparql.algebra.Table;
import com.hp.hpl.jena.sparql.algebra.TableFactory;
import com.hp.hpl.jena.sparql.algebra.op.OpBGP;
import com.hp.hpl.jena.sparql.algebra.op.OpLeftJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpProject;
import com.hp.hpl.jena.sparql.engine.ref.Evaluator;

public class EvalExt {

	public static boolean debug = false;
	private final Evaluator evaluator;

	public EvalExt(Evaluator ev) {
		this.evaluator = ev;
	}

	/**
	 * Evaluates the given tree pattern with the provided evaluator
	 *
	 * @param tree
	 * @return
	 */
	public Table eval(TreePattern tree) {
		return ext(TableFactory.createUnit(), tree.getRoot());
	}

	private Table ext(Table M, TreePatternNode n) {
		Table res = null;
		if (n.getBgp() != null) {
			// n is a BGP node
			long begin_time = System.currentTimeMillis();
			res = evaluator.basicPattern(n.getBgp().getPattern());
			if (n.getFilter() != null) {
				res = evaluator.filter(n.getFilter(), res);
			}
			long end_time = System.currentTimeMillis();
			n.setEvaluationTime(end_time - begin_time);
			n.setNumberOfLocalMappings(res.size());
		} else {
			// node is a union node
			Table M_union = TableFactory.create();
			for (TreePatternNode n_i : n.getChildren()) {
				res = TableFactory.createUnit();
				Table res_i = ext(res, n_i);
				M_union = evaluator.union(M_union, res_i);
			}
			return M_union;
		}

		if (M.isEmpty()) {
			return M;
		}

		res = evaluator.join(M, res);
		n.setNumberOfResults(res.size());
		if (n.getChildren().isEmpty()) {
			// n is a leaf node
			return res;
		}

		Table M_join = TableFactory.createUnit();
		for (TreePatternNode n_i : n.getChildren()) {
			// we cannot be sure what the child does with our mappings so we have to copy them
			Table res_parent = TableFactory.create(res.iterator(evaluator.getExecContext()));
			Table res_i = ext(res_parent, n_i);
			res_parent = TableFactory.create(res.iterator(evaluator.getExecContext()));
			Table M_i = evaluator.leftJoin(res_parent, res_i, null);
			// if(M_i.isEmpty()) //this should never be empty if res is not
			// empty
			// return M_i;
			M_join = evaluator.join(M_join, M_i);
		}
		return M_join;
	}

	public Table eval(Table M, Op op) {
		if (!WellDesignedService.isWellDesigned(op)) {
			System.err.println("Warning, non well-designed query. Proceeding anyway.");
		}
		if (op instanceof OpBGP) {
			Table res = evaluator.basicPattern(((OpBGP) op).getPattern());
			return evaluator.join(M, res);
		}
		if (op instanceof OpLeftJoin) {
			TreePattern tree;
			try {
				tree = TreePatternFactory.create(op);
			} catch (UnsupportedOperatorForPatternTreeException | PatternTreeFactoryException e) {
				fail(op);
				return null;
			}
			TreePatternNode root = tree.getRoot();

			if (debug) {
				root.dump();
			}
			return ext(M, root);
		}
		if (op instanceof OpProject) {
			Table res = eval(M, ((OpProject) op).getSubOp());
			return evaluator.project(res, ((OpProject) op).getVars());
		}

		fail(op);
		return null;
	}

	private static void fail(Op op) {
		System.err.println("Error: unsupported operation: " + op.getName());
		System.err.println(op);
		System.exit(1);
	}
}
