/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package spam.gui.graph;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.Rectangle2D;
import java.text.AttributedString;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JOptionPane;

import spam.model.SubtreeMatch;
import spam.model.TreePattern.WdConflict;
import spam.model.TreePatternNode;
import spam.transformationRules.R1Application;
import spam.transformationRules.R2Application;
import spam.transformationRules.R3Application;
import spam.transformationRules.R4Application;
import spam.transformationRules.RuleApplication;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.shared.PrefixMapping;
import com.hp.hpl.jena.sparql.expr.Expr;
import com.hp.hpl.jena.sparql.expr.ExprList;

/**
 *
 * @author Andres Letelier
 */
public class GraphicalNode extends JComponent {

	private static final long serialVersionUID = 1L;
	private static final boolean CENTERED = true;
	private static final Color HIGHLIGHT_COLOR = Color.RED;
	private static final Color RULE_COLOR = Color.BLUE;
	private static final Color REDUNDANT_COLOR = Color.GRAY;
	private static final int borderTop = 5, borderBottom = 5, borderLeft = 5, borderRight = 5, siblingDistance = 10, borderGap = 2, levelDistance = 10;
	private TreePatternNode treeNode;
	private String highlightVariable = null;
	private RuleApplication ruleApplication = null;
	private boolean highlightNode = false;
	private boolean counterexampleNode = false;
	private boolean viewportMode = false;
	private int maxWidth = 0;
	private int outerWidth = 0;
	private Rectangle nodeBox = null;
	private Rectangle highlightBox = null;
	private final List<GraphicalNode> children = new LinkedList<GraphicalNode>();
	private PrefixMapping prefixes = null;

	public TreePatternNode getNode() {
		return treeNode;
	}

	/**
	 * clears all highlighting from this node downwards
	 */
	public void removeHighlight() {
		this.highlightNode = false;
		this.highlightVariable = null;
		for (GraphicalNode child : children) {
			child.removeHighlight();
		}
	}

	public void setNode(TreePatternNode node) {
		this.treeNode = node;
	}

	/**
	 * this is where the magic happens
	 *
	 * @param g
	 */
	@Override
	public void paintComponent(Graphics g) {
		if (g instanceof Graphics2D) {
			this.paintNode((Graphics2D) g, 0, 0, false);
		}

	}

	@Override
	public void paint(Graphics g) {
		this.paintComponent(g);
	}

	/**
	 * returns the value of outerWidth as computed by the last run of recomputeMaxWidth.
	 *
	 * Care! This value may be outdated, since no care is taken to update this value if the shape of the tree changes. It is in the responsibility of the caller to make sure that no change occured
	 * since the last call of recomputeOuterWidth
	 *
	 * The outerwidth of a node is the maximium of its maxWidth and the sum of the outerwidth of its children plus some space between those children.
	 *
	 * @return the value computed during the last run of recomputeOuterWidth
	 */
	public int getOuterWidth() {
		return this.outerWidth;
	}

	public int recomputeOuterWidth(Graphics g) {

		this.recomputeMaxWidth(g);

		int childWidth = 0;
		for (GraphicalNode c : this.children) {
			if (childWidth > 0) {
				childWidth += siblingDistance;
			}
			childWidth += c.recomputeOuterWidth(g);
		}

		int ownWidth = this.getMaxWidth() + borderLeft + borderRight + 5;

		if (ownWidth > childWidth) {
			this.outerWidth = ownWidth;
		} else {
			this.outerWidth = childWidth;
		}

		return this.outerWidth;
	}

	/**
	 * returns the value of maxWidth as computed by the last run of recomputeMaxWidth.
	 *
	 * Care! This value may be outdated, since no care is taken to update this value if the shape of the tree changes. It is in the responsibility of the caller to make sure that no change occurred
	 * since the last call of recomputeMaxWidth
	 *
	 * @return the value computed during the last run of recomputeMaxWidth
	 */
	public int getMaxWidth() {
		return this.maxWidth;
	}

	public int recomputeMaxWidth(Graphics g) {
		FontMetrics fm = g.getFontMetrics();
		List<AttributedString> lines = this.getAttributedLines();

		// node content
		this.maxWidth = 0;
		for (AttributedString as : lines) {
			if (getAttributedTextWidth(g, as) > this.maxWidth) {
				this.maxWidth = getAttributedTextWidth(g, as);
			}
		}

		for (String filter : getFilters()) {
			int filterWidth = fm.stringWidth(filter);
			if (filterWidth > this.maxWidth) {
				this.maxWidth = filterWidth;
			}
		}

		// node statistics
		if (this.displayStatistics()) {
			int statswidth = fm.stringWidth(this.getStatisticString());
			if (statswidth > this.maxWidth) {
				this.maxWidth = statswidth;
			}
		}

		return this.maxWidth;
	}

	private int getAttributedTextWidth(Graphics g, AttributedString text) {
		TextLayout textLayout = new TextLayout(text.getIterator(), ((Graphics2D) g).getFontRenderContext());
		Rectangle2D textBounds = textLayout.getBounds();
		return (int) textBounds.getMaxX();
	}

	public int getMaxHeight(Graphics g) {

		// first get the max Height from the children
		int y = 0;

		if (this.children != null) {

			for (GraphicalNode c : this.children) {
				int ch = c.getMaxHeight(g);
				if (ch > y) {
					y = ch;
				}
			}

			y += levelDistance;
		}

		// then add the own height
		FontMetrics fm = g.getFontMetrics();
		List<AttributedString> lines = this.getAttributedLines();
		List<String> filters = this.getFilters();

		if (this.displayStatistics()) {
			y += borderTop + fm.getAscent() + fm.getHeight();
		}

		y += borderTop + fm.getAscent();
		y += fm.getHeight() * lines.size();
		y += fm.getHeight() * filters.size();
		y += borderBottom;

		return y;
	}

	private String parseExpression(Node node) {

		if (node.isURI()) {
			return node.toString(this.prefixes);
		}
		if (node.isVariable()) {
			return node.toString();
		}
		if (node.isLiteral()) {
			return node.toString();
		}
		return "";
	}

	/**
	 * Returns the color that shall be used to write the given position in a triple pattern.
	 *
	 * The position of the triple pattern is specified via the argument. At the moment, the following colors apply: black as default, and read either if the complete node shall be highlighted or the
	 * specified pattern is a variable that shall be highlighted.
	 *
	 * @param node
	 *            the position of a triple pattern to be colored
	 * @return red if highlighted, black as default
	 */
	private Color getNodeColor(Node node, Triple t) {

		if (this.treeNode.isRedundantTriple(t)) {
			return REDUNDANT_COLOR;
		}

		if (this.ruleApplication != null) {

			if (this.ruleApplication instanceof R1Application) {
				R1Application r = (R1Application) this.ruleApplication;
				if (r.getDeleteTriples().contains(t)) {
					return RULE_COLOR;
				}
			}

			if (this.ruleApplication instanceof R2Application) {
				return RULE_COLOR;
			}

			if (this.ruleApplication instanceof R3Application) {
				return RULE_COLOR;
			}

			if (this.ruleApplication instanceof R4Application) {
				return RULE_COLOR;
			}
		}

		if (this.highlightNode) {
			return HIGHLIGHT_COLOR;
		}

		// default highlight
		if (node.isVariable() && node.getName().equals(this.highlightVariable)) {
			return Color.red;
		}
		return Color.black;
	}

	/**
	 * returns the color that shall be used to draw the surrounding rectangle of the node.
	 *
	 * At the moment, the default color is black, and if the node contains a variable that should be highlighted, it is red.
	 *
	 * @return
	 */
	private Color getBoxColor(boolean highlightCounterexample) {

		if (this.ruleApplication != null) {
			return RULE_COLOR;
		}

		if (this.treeNode.getVariableNamesIncludingFilter().contains(this.highlightVariable)) {
			return HIGHLIGHT_COLOR;
		}

		if (this.counterexampleNode && highlightCounterexample) {

			return HIGHLIGHT_COLOR;
		}

		return Color.black;
	}

	/**
	 * Returns if this node is highlighted or not.
	 * 
	 * Used e.g. by the parent in order to determine which color or form the edge shall have.
	 * 
	 * @param highlightCounterexample
	 * @return
	 */
	boolean nodeIsHighlighted(boolean highlightCounterexample) {

		if (this.ruleApplication != null) {
			return true;
		}

		if (this.treeNode.getVariableNamesIncludingFilter().contains(this.highlightVariable)) {
			return true;
		}

		if (this.counterexampleNode && highlightCounterexample) {
			return true;
		}

		return false;

	}

	/**
	 * returns the color of the edge connecting this node to the specified child.
	 * 
	 * Returns HIGHLIGHT_COLOR if for both nodes the method nodeIsHighlighted(highlightCounterxample) returns true, black otherwise.
	 *
	 * @param toChild
	 *            the child to which the outgoing edge leads
	 * @return
	 */
	private Color getEdgeColor(GraphicalNode toChild, boolean highlightCounterexample) {

		if (this.nodeIsHighlighted(highlightCounterexample) && toChild.nodeIsHighlighted(highlightCounterexample)) {
			return HIGHLIGHT_COLOR;
		}

		return Color.black;
	}

	private List<AttributedString> getAttributedLines() {
		if (treeNode.getBgp() == null) {
			List<AttributedString> line = new ArrayList<>();
			line.add(new AttributedString("UNION"));
			return line;
		}
		List<AttributedString> aLines = new ArrayList<>();

		for (Triple t : treeNode.getBgp().getPattern()) {

			StringBuilder sb = new StringBuilder();
			int sEnd;
			int pEnd;

			sb.append(this.parseExpression(t.getSubject()));
			sEnd = sb.length();

			sb.append(" ");
			sb.append(this.parseExpression(t.getPredicate()));
			pEnd = sb.length();

			sb.append(" ");
			sb.append(this.parseExpression(t.getObject()));

			AttributedString aString = new AttributedString(sb.toString());
			// for subject
			aString.addAttribute(TextAttribute.FOREGROUND, getNodeColor(t.getSubject(), t), 0, sEnd);

			// for predicate
			aString.addAttribute(TextAttribute.FOREGROUND, getNodeColor(t.getPredicate(), t), sEnd + 1, pEnd);

			// for object
			aString.addAttribute(TextAttribute.FOREGROUND, getNodeColor(t.getObject(), t), pEnd + 1, sb.length());

			aLines.add(aString);
		}

		return aLines;
	}

	private List<String> getFilters() {
		ArrayList<String> filters = new ArrayList<>();
		ExprList exprList = this.treeNode.getFilter();
		if (exprList == null) {
			return filters;
		}
		for (Expr expr : exprList) {
			filters.add(expr.toString());
		}
		return filters;
	}

	public void paintNode(Graphics2D g, int offsetX, int offsetY, boolean highlightCounterexample) {

		FontMetrics fm = g.getFontMetrics();
		List<AttributedString> lines = this.getAttributedLines();
		List<String> filters = this.getFilters();

		// to center each line we need to figure out which line is the widest
		this.recomputeMaxWidth(g);

		int nOffsetX = offsetX + this.outerWidth / 2 - this.getMaxWidth() / 2;
		int nOffsetY = offsetY;

		int y = borderTop + fm.getAscent();

		// check if we have to display statistical information
		if (this.displayStatistics()) {

			String stats = this.getStatisticString();
			g.drawString(stats, nOffsetX + borderLeft + (maxWidth - fm.stringWidth(stats)) / 2, nOffsetY + y);
			y += borderTop + fm.getAscent() + fm.getHeight();
		}

		// now we actually draw the lines centered if GraphicalNode.CENTERED is true

		for (AttributedString as : lines) {
			if (CENTERED) {
				g.drawString(as.getIterator(), nOffsetX + borderLeft + (maxWidth - getAttributedTextWidth(g, as)) / 2, nOffsetY + y);
			} else {
				g.drawString(as.getIterator(), nOffsetX + borderLeft, nOffsetY + y);
			}
			y += fm.getHeight();
		}

		for (String filter : filters) {
			if (CENTERED) {
				g.drawString(filter, nOffsetX + borderLeft + (maxWidth - fm.stringWidth(filter)) / 2, nOffsetY + y);
			} else {
				g.drawString(filter, nOffsetX + borderLeft, nOffsetY + y);
			}
			y += fm.getHeight();
		}

		g.setColor(this.getBoxColor(highlightCounterexample));
		if (this.nodeIsHighlighted(highlightCounterexample)) {
			this.nodeBox = new Rectangle(nOffsetX, nOffsetY, borderLeft + maxWidth + borderRight, y + borderBottom - fm.getHeight());
			g.draw(this.nodeBox);

			this.highlightBox = new Rectangle(nOffsetX + borderGap, nOffsetY + borderGap, borderLeft + maxWidth + borderRight - 2 * borderGap, y + borderBottom - fm.getHeight() - 2 * borderGap);
			g.draw(this.highlightBox);

		} else {
			this.nodeBox = new Rectangle(nOffsetX, nOffsetY, borderLeft + maxWidth + borderRight, y + borderBottom - fm.getHeight());
			g.draw(this.nodeBox);
		}
		g.setColor(Color.BLACK);

		int outEdgePadding = 0;
		if (this.children.size() > 0) {
			outEdgePadding = this.getMaxWidth() / this.children.size();
		}
		int nextEdgeStart = nOffsetX + borderLeft + outEdgePadding / 2;

		// Now paint the children
		int childOffsetX = offsetX;
		int childOffsetY = nOffsetY + y + borderBottom + levelDistance;

		// if the width of the children is smaller than of the parent,
		// shift the children by half of the difference (i.e. center the
		// children)
		int childrenOuterWidth = 0;
		for (GraphicalNode child : this.children) {
			if (childrenOuterWidth > 0) {
				childrenOuterWidth += siblingDistance;
			}
			childrenOuterWidth += child.getOuterWidth();
		}

		if (childrenOuterWidth < this.getOuterWidth()) {
			childOffsetX += Math.ceil((this.getOuterWidth() - childrenOuterWidth) / 2.00);
		}

		for (GraphicalNode c : this.children) {
			c.paintNode(g, childOffsetX, childOffsetY, highlightCounterexample);

			// draw the line to the child node
			g.setColor(this.getEdgeColor(c, highlightCounterexample));

			if (this.nodeIsHighlighted(highlightCounterexample) && c.nodeIsHighlighted(highlightCounterexample)) {
				// highlight edge
				g.drawLine(nextEdgeStart - 1, nOffsetY + y + borderBottom - fm.getHeight(), (childOffsetX + c.getOuterWidth() / 2 + borderLeft) - 1, childOffsetY);
				g.drawLine(nextEdgeStart + 1, nOffsetY + y + borderBottom - fm.getHeight(), (childOffsetX + c.getOuterWidth() / 2 + borderLeft) + 1, childOffsetY);
			} else {
				// normal edge
				g.drawLine(nextEdgeStart, nOffsetY + y + borderBottom - fm.getHeight(), childOffsetX + c.getOuterWidth() / 2 + borderLeft, childOffsetY);
			}

			nextEdgeStart += outEdgePadding;

			childOffsetX += c.getOuterWidth() + siblingDistance;
		}

	}

	public void paintDragImage(Graphics2D g2d, int xOffset, int yOffset) {

		// maybe we want something special for this one, but for the
		// moment
		// lets keep it siple
		Rectangle oldNodeBox = this.nodeBox;

		this.paintNode(g2d, xOffset - this.outerWidth / 2, yOffset, false);

		this.nodeBox = oldNodeBox;

		/*
		 * The fact that we do it only here and not in the child nodes as well helps us to avoid to drop the node onto one of its children.
		 */
	}

	public SubtreeMatch isSubtree(TreePatternNode pattern) {

		if (!this.treeNode.equalsBGP(pattern)) {
			return null;
		}

		// the first problem is settled, up to the hard part

		List<List<SubtreeMatch>> possibleMatches = new LinkedList<List<SubtreeMatch>>();

		// first of all we find possible matchings for each child of the
		// subtree-root.
		for (TreePatternNode patternChild : pattern.getChildren()) {

			List<SubtreeMatch> curMatches = new LinkedList<SubtreeMatch>();

			for (GraphicalNode child : this.children) {

				SubtreeMatch childMatches = child.isSubtree(patternChild);

				if (childMatches != null) {
					// a candidate matching for this child
					curMatches.add(childMatches);
				}
			}

			if (curMatches.size() <= 0) {
				// we could not find a match for this pattern
				return null;
			}

			possibleMatches.add(curMatches);
		}

		// Now we have to find a one-to-one matching between possible partners.
		// we leave this open for the moment

		// collect the results
		List<SubtreeMatch> results = new LinkedList<SubtreeMatch>();
		for (List<SubtreeMatch> curmatch : possibleMatches) {
			results.add(curmatch.get(0));
		}

		// and return how to match the pattern into this tree
		return new SubtreeMatch(pattern, this, results);
	}

	/**
	 * Sets the flag indicating that this node is part of a counterexample to true.
	 * 
	 * not more.
	 */
	public void setCounterexample() {
		this.counterexampleNode = true;
	}

	/**
	 * marks the subtree rooted at this node as not part of a counterexample.
	 *
	 * Sets the counterexampleNode flag to false and recursively calls the same function on all children.
	 */
	public void unsetCounterexample() {
		this.counterexampleNode = false;
		for (GraphicalNode child : this.children) {
			child.unsetCounterexample();
		}
	}

	public void init() {
		this.children.clear();

		for (TreePatternNode c : this.treeNode.getChildren()) {
			GraphicalNode newChild = new GraphicalNode(c);
			newChild.init();
			this.children.add(newChild);
		}
	}

	public void setWdHighlight(WdConflict conflict) {

		if (conflict.isConflictNode(this.treeNode)) {
			this.highlightVariable = conflict.getVarname();
		} else {
			this.highlightVariable = null;
		}

		for (GraphicalNode c : this.children) {
			c.setWdHighlight(conflict);
		}
	}

	public void setHighlightRuleApplicationNode(RuleApplication application) {

		if (this.treeNode == application.getHighlightNode()) {
			this.ruleApplication = application;
		} else {
			this.ruleApplication = null;
		}

		for (GraphicalNode c : this.children) {
			c.setHighlightRuleApplicationNode(application);
		}
	}

	public GraphicalNode(TreePatternNode node) {
		this.treeNode = node;

		// this.addMouseListener(new MouseHandler(null));
	}

	/**
	 * checks if the given point lies within the box of this node.
	 *
	 * if yes, this node is returned as the target, otherwise the request is handed further to the children (YES, this COULD be implement much more efficiently, but at the moment we do an exhaustive
	 * search)
	 *
	 * @param clickPoint
	 *            the point to checkFor
	 * @return
	 */
	public GraphicalNode mouseInside(Point clickPoint) {

		// the mouse pointer was indeed over this box
		if (this.nodeBox != null && this.nodeBox.contains(clickPoint)) {
			return this;
		}

		// no, the mouse pointer was not over this element
		for (GraphicalNode c : this.children) {
			GraphicalNode result = c.mouseInside(clickPoint);
			if (result != null) {
				return result;
			}
		}

		return null;

		// note that one could search much more efficiently
		// for the right component ... after all, we have a tree here!!
		// (binary search ftw!!
		// (however, I first have to check which coordindate translations
		// are necessary to make actually use of it)

	}

	private String formatForEditing(Node node, PrefixMapping prefixes) {

		if (node.isURI() && node.toString().equals(node.toString(prefixes))) {
			return "<" + node.toString() + ">";
		}

		return node.toString(prefixes);
	}

	public boolean editContent(PrefixMapping prefixes) {

		StringBuilder sb = new StringBuilder();
		// System.out.println(this.treeNode.toString());

		for (Triple t : this.treeNode.getBgp().getPattern()) {

			if (sb.length() > 0) {
				sb.append(" . ");
			}
			sb.append(this.formatForEditing(t.getSubject(), prefixes));
			sb.append(" ");
			sb.append(this.formatForEditing(t.getPredicate(), prefixes));
			sb.append(" ");
			sb.append(this.formatForEditing(t.getObject(), prefixes));
		}

		String newContent = JOptionPane.showInputDialog(
		// this,
				"Enter the node content",
				// "Edit graph pattern",
				sb.toString());

		if (newContent != null) {

			try {

				if (!this.treeNode.setBgp(newContent, this.prefixes)) {
					JOptionPane.showMessageDialog(null, "Please provide a valid BGP" + "in Turtle syntax");
					return false;
				} else {
					return true;
				}
			} catch (Exception e) {
				JOptionPane.showMessageDialog(null, "Cannot parse the input: " + e.toString());
				return false;
			}
		}

		return false;
	}

	void delete(int mode) {

		if (mode == 0) {
			this.treeNode.delete();
		} else if (mode == 1) {
			this.treeNode.deleteSubtree();
		}
	}

	void addChild(GraphicalNode child) {
		this.children.add(child);
	}

	public void setPrefixes(PrefixMapping prefixes) {
		this.prefixes = prefixes;

		for (GraphicalNode c : this.children) {
			c.setPrefixes(prefixes);
		}
	}

	public void setStatistics(int bindings, long runtime) {
		this.setNumbVariableBindings(bindings);
		this.setRuntime(runtime);
	}

	public void setNumbVariableBindings(int numb) {
		treeNode.setNumberOfResults(numb);
	}

	public void setRuntime(long runtime) {
		treeNode.setEvaluationTime(runtime);
	}

	public void unsetStatistics() {

		// unrealistic values are used to indicate that it shall not be
		// displayed
		this.setStatistics(-1, -1);

		// propagate this to the children
		for (GraphicalNode c : this.children) {
			c.unsetStatistics();
		}
	}

	private boolean displayStatistics() {

		if (this.viewportMode) {
			return false;
		}

		return treeNode.getEvaluationTime() >= 0 || treeNode.getNumberOfResults() >= 0;
	}

	void setViewportMode(boolean active) {
		this.viewportMode = active;

		for (GraphicalNode c : this.children) {
			c.setViewportMode(active);
		}
	}

	private String getStatisticString() {
		return "runtime: " + treeNode.getEvaluationTime() + "ms" + "   #var. bindings: " + treeNode.getNumberOfLocalMappings() + "   #results: " + treeNode.getNumberOfResults();
	}
}