package spam.model;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.List;

import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import spam.exception.PatternTreeFactoryException;
import spam.gui.QueryListHandler;
import spam.gui.RuleApplicationList;
import spam.gui.graph.GraphicalTree;
import spam.model.TreePattern.WdConflict;
import spam.transformationRules.RuleApplication;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.shared.PrefixMapping;
import com.hp.hpl.jena.sparql.algebra.OpAsQuery;
import com.hp.hpl.jena.sparql.algebra.Table;

/**
 * Manages a query and can be displayed graphically
 *
 * @author Andres Letelier
 */
public class QueryComponent extends JPanel {

	private static final long serialVersionUID = 1L;
	private ResultTableModel resultTableModel;
	private Query query;
	private String name;
	private TreePattern tp;
	private GraphicalTree graphicalTree;
	private String unsavedQueryString = null;
	private TreeViewer viewport1;
	private TreeViewer viewport2;
	private File queryFile = null;
	JScrollPane component;
	private String statisticsString;

	public void setResultTableModel(Table table) {
		this.resultTableModel = new ResultTableModel(table, query.getPrefixMapping());
	}

	public TableModel getResultTableModel() {
		return this.resultTableModel;
	}

	public String getStatisticsString() {
		return statisticsString;
	}

	public void setStatisticsString(String statisticsString) {
		this.statisticsString = statisticsString;
	}

	GraphicalTree viewport1Tree, viewport2Tree;

	public GraphicalTree getGraphicalTree() {
		return graphicalTree;
	}

	public TreePattern getTreePattern() {

		if (tp != null) {
			return tp;
		}
		try {
			return TreePatternFactory.create(query);
		} catch (PatternTreeFactoryException e) {
			return null;
		}
	}

	public RuleApplicationList getRuleApplications() {
		return new RuleApplicationList(this);
	}

	public void setRuleApplications(RuleApplicationList r) {
	}

	public void updateGraph() {
		this.graphicalTree.init(this.tp);
		this.graphicalTree.repaint();
	}

	/**
	 * Used to populate the graph recursively from the TreePattern
	 *
	 * @param n
	 */
	/*
	 * private void graphBuilder(TreePatternNode n) { Object parent=graph.getDefaultParent(); Object ncell=((mxGraphModel)graph.getModel()).getCell(n.getId());
	 * 
	 * for (TreePatternNode c : n.getChildren()) { Object ccell=graph.insertVertex(parent,c.getId(),prettifyBGP(c.toString()), 0, 0, DEFAULT_CELL_WIDTH, DEFAULT_CELL_HEIGHT); graph.insertEdge(parent,
	 * n.getId()+c.getId(), "", ncell, ccell); graph.updateCellSize(ccell); graphBuilder(c); } }
	 */
	@Override
	public String toString() {
		return name;
	}

	public void updateViewportView() {
		component.setViewportView(graphicalTree);
		component.repaint();
		component.revalidate();
	}

	private void init() {

		this.component = new JScrollPane();
		this.component.setBackground(Color.white);
		BoxLayout layout = new BoxLayout(this, BoxLayout.Y_AXIS);
		this.setLayout(layout);
		component.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		component.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);

		graphicalTree = new GraphicalTree(this);
		component.setViewportView(graphicalTree);

		this.updateGraph();
		this.add(component);
	}

	public QueryComponent(String name, Query query) {
		super();
		this.query = query;
		this.name = name;
		try {
			this.tp = TreePatternFactory.create(query);
		} catch (PatternTreeFactoryException e) {
			this.tp = null;
		} finally {
			init();
		}
	}

	public QueryComponent(String name, TreePattern tree) {
		super();
		this.tp = tree;
		this.name = name;
		this.query = OpAsQuery.asQuery(tree.toOP());
		init();
	}

	public QueryComponent(String name, Query query, File file) {
		this(name, query);
		this.queryFile = file;
	}

	public void setQueryFile(File file) {
		this.queryFile = file;
	}

	public File getQueryFile() {
		return this.queryFile;
	}

	public boolean saveQueryToFile(File targetFile) {
		if (targetFile == null) {
			return false;
		}
		this.queryFile = targetFile;

		try {
			// Create file
			FileWriter fstream = new FileWriter(queryFile);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(query.serialize());
			// Close the output stream
			out.close();
		} catch (Exception e) {// Catch exception if any
			JOptionPane.showMessageDialog(null, "Error saving to file: " + e.getMessage());
			return false;
		}

		return true;
	}

	public Query getQuery() {
		return query;
	}

	public String getQueryName() {
		return name;
	}

	public void setQueryName(String name) {
		this.name = name;
	}

	public void setQuery(Query query) throws PatternTreeFactoryException {
		this.query = query;
		this.syncFromQuery();
	}

	public void setPrefixMapping(PrefixMapping prefixes) {
		this.query.setPrefixMapping(prefixes);
	}

	public void syncFromGraph() {
		PrefixMapping pm = this.query.getPrefixMapping();
		this.query = OpAsQuery.asQuery(this.tp.toOP());
		this.query.setPrefixMapping(pm);
		QueryListHandler.getInstance().queryGraphUpdated(this);

		QueryListHandler.getInstance().queryUpdated(this);
		this.updateGraph();
	}

	/**
	 * the query must be in opt-normal form at this point
	 * 
	 * @throws PatternTreeFactoryException
	 */
	public void syncFromQuery() throws PatternTreeFactoryException {
		try {
			this.tp = TreePatternFactory.create(query);
		} catch (PatternTreeFactoryException e) {
			this.tp = null;
			this.updateGraph();
			throw e;
		}
		QueryListHandler.getInstance().queryUpdated(this);
		this.updateGraph();
	}

	public boolean hasUnsavedChanges() {
		return this.unsavedQueryString != null;
	}

	public void setUnsavedQueryString(String query) {
		this.unsavedQueryString = query;
	}

	public String getUnsavedQueryString() {
		return this.unsavedQueryString;
	}

	public void querySaved() {
		this.unsavedQueryString = null;
	}

	public void setRuleApplicationNode(RuleApplication application) {
		this.graphicalTree.setHighlightRuleApplicationNode(application);

	}

	public void setWdConflictHighlightVar(WdConflict conflict) {
		this.graphicalTree.setHighlightVar(conflict);
	}

	public boolean isQuasiWellDesigned() {
		if (tp == null) {
			this.getTreePattern();
		}
		tp.checkQuasiWellDesigned();
		return tp.isQuasiWellDesigned();
	}

	public boolean isWellDesigned() {
		if (tp == null) {
			this.getTreePattern();
		}
		tp.checkWellDesigned();
		return tp.isWellDesigned();
	}

	public void setCounterexampleSubtree(TreePattern subtree) {
		this.graphicalTree.highlightCounterexample(subtree);
	}

	public void unsetCounterexampleSubtree() {
		this.graphicalTree.unsetCounterexample();
	}

	public JComponent getViewport1() {
		if (this.viewport1 == null) {
			this.viewport1 = new TreeViewer(this.graphicalTree);
		}
		return this.viewport1;
	}

	public JComponent getViewport2() {
		if (this.viewport2 == null) {
			this.viewport2 = new TreeViewer(this.graphicalTree);
		}
		return this.viewport2;
	}

	public void repaintViewports() {

		this.getViewport1().repaint();
		this.getViewport1().revalidate();

		this.getViewport2().repaint();
		this.getViewport2().revalidate();

	}

	public class TreeViewer extends JPanel {

		private static final long serialVersionUID = 1L;
		GraphicalTree tree;

		public TreeViewer(GraphicalTree parent) {

			super();

			this.tree = parent;
			setBackground(Color.white);
			this.add(parent);
		}

		@Override
		public void paintComponent(Graphics g) {

			this.tree.paintWithCounterexamples(g);
			Dimension size = this.tree.getPreferredSize();

			/*
			 * if (this.getParent().getHeight() > size.height) { size = new Dimension(size.width, this.getParent().getHeight()); this.tree.setPreferredSize(size); this.tree.revalidate(); }
			 */
			this.setPreferredSize(size);
			this.revalidate();
		}
	}

	public class ResultTableModel extends AbstractTableModel {

		private static final long serialVersionUID = 1L;
		Table results;
		List<String> cols;
		String[][] data;
		private boolean tooManyResults;

		public ResultTableModel(Table table, PrefixMapping prefixes) {

			cols = table.getVarNames();
			data = new String[table.size()][cols.size()];
			ResultSet res = table.toResultSet();

			try {
				while (res.hasNext()) {
					QuerySolution sol = res.next();
					for (int i = 0; i < cols.size(); i++) {
						if (sol.get(cols.get(i)) != null) {
							data[res.getRowNumber() - 1][i] = sol.get(cols.get(i)).asNode().toString(prefixes);
						} else {
							// partial assignment that does not bind this
							// variable
							data[res.getRowNumber() - 1][i] = "";
						}
					}
				}
				this.tooManyResults = false;
			} catch (ArrayIndexOutOfBoundsException aioobe) {
				this.tooManyResults = true;
			}
		}

		public boolean containsAllResults() {
			return !this.tooManyResults;
		}

		@Override
		public int getRowCount() {
			return data.length;
		}

		@Override
		public int getColumnCount() {
			return cols.size();
		}

		@Override
		public Object getValueAt(int row, int col) {
			return data[row][col];
		}

		@Override
		public String getColumnName(int i) {
			return cols.get(i);
		}

	}
}
