package spam.model;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.hp.hpl.jena.sparql.algebra.Op;
import com.hp.hpl.jena.sparql.algebra.op.OpBGP;
import com.hp.hpl.jena.sparql.algebra.op.OpDistinct;
import com.hp.hpl.jena.sparql.algebra.op.OpFilter;
import com.hp.hpl.jena.sparql.algebra.op.OpJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpLeftJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpOrder;
import com.hp.hpl.jena.sparql.algebra.op.OpProject;
import com.hp.hpl.jena.sparql.algebra.op.OpReduced;
import com.hp.hpl.jena.sparql.algebra.op.OpSlice;
import com.hp.hpl.jena.sparql.algebra.op.OpUnion;

/**
 * Wrapper to keep people from messing with nodes and guessing where the root is. Don't touch my nodes!
 * 
 * @author Andres Letelier
 */
public class TreePattern {

	/*
	 * There might be a bunch of modifiers on top of my beautiful pattern, like projections, slices and whatnot. These nasty modifiers must be incarcerated and eventually executed (meaning that I keep
	 * them sepparately in the head variable).
	 */
	private Op head;

	public Op getHead() {
		return head;
	}

	/**
	 * Deep copy of the tree body.
	 * 
	 * @return
	 */
	@Override
	public TreePattern clone() {
		TreePattern copy = new TreePattern();
		copy.head = this.head;
		copy.qwdCounterExamples = this.qwdCounterExamples;
		copy.root = this.root.deepCopy();
		copy.checkQuasiWellDesigned();
		copy.checkWellDesigned();
		return copy;
	}

	/**
	 * Returns the Op form of the tree, wrapped with the modifiers from the head.
	 * 
	 * @return
	 */
	public Op toOP() {
		return toOpWorker(head);
	}

	/**
	 * Returns the Op form of the tree ignoring the head
	 * 
	 * @return
	 */
	public Op toOPNoHead() {
		return root.toOp();
	}

	/**
	 * Dive into the appropriate modifiers until we find a valid tree pattern op (meaning left join or basic pattern or Union)
	 * 
	 * @param head
	 * @return
	 */
	private Op toOpWorker(Op head) {
		if (isAcceptedBodyOp(head)) {
			return root.toOp();
		}
		if (head instanceof OpProject) {
			OpProject op = (OpProject) head;
			return new OpProject(toOpWorker(op.getSubOp()), op.getVars());
		}
		if (head instanceof OpSlice) {
			OpSlice op = (OpSlice) head;
			return new OpSlice(toOpWorker(op.getSubOp()), op.getStart(), op.getLength());
		}
		if (head instanceof OpReduced) {
			OpReduced op = (OpReduced) head;
			return OpReduced.create(toOpWorker(op.getSubOp()));
		}
		if (head instanceof OpOrder) {
			OpOrder op = (OpOrder) head;
			return new OpOrder(toOpWorker(op.getSubOp()), op.getConditions());
		}
		if (head instanceof OpDistinct) {
			OpDistinct op = (OpDistinct) head;
			return new OpDistinct(toOpWorker(op.getSubOp()));
		}
		throw new UnsupportedOperationException(head.getClass().getSimpleName());
	}

	/**
	 * Keep track of handled operators
	 * 
	 * @param op
	 * @return true if op instanceof [OpProject, OpSlice, OpReduced, OpOrder, OpDistinct]
	 */
	public static boolean isAcceptedHeadOp(Op op) {
		return op instanceof OpProject || op instanceof OpSlice || op instanceof OpReduced || op instanceof OpOrder || op instanceof OpDistinct;
	}

	/**
	 * Keep track of handled operators
	 * 
	 * @param op
	 * @return true if op instanceof [OpBGP, OpJoin, OpLeftJoin, OpFilter, OpUnion]
	 */
	public static boolean isAcceptedBodyOp(Op op) {
		return op instanceof OpBGP || op instanceof OpJoin || op instanceof OpLeftJoin || op instanceof OpFilter || op instanceof OpUnion;
	}

	public void setHead(Op head) {
		this.head = head;
	}

	private TreePatternNode root;
	private QwdStatus qwdCounterExamples;
	private QwdStatus wdCounterExamples;

	public TreePattern(TreePatternNode root) {
		this.root = root;
	}

	public TreePattern() {
		this.root = null;
	}

	public TreePatternNode getRoot() {
		return root;
	}

	public void setRoot(TreePatternNode root) {
		this.root = root;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (!(o instanceof TreePattern)) {
			return false;
		}
		TreePattern other = (TreePattern) o;
		return this.root.equals(other.root);
	}

	@Override
	public int hashCode() {
		int hash = 3;
		hash = 17 * hash + (this.root != null ? this.root.hashCode() : 0);
		return hash;
	}

	public void dump() {
		this.root.dump();
	}

	/**
	 * very (extremely) similar to what dump outputs - sorry for the doubling
	 * 
	 * @return
	 */
	public String getStringRepresentation() {
		return this.root.getDebugString(1).toString();
	}

	public void setStatusString(String status) {
	}

	/**
	 * checks if the pattern tree is well-designed.
	 * 
	 * if the tree consists of no node, then true is returned. Otherwise it is checked if the connectedness condition holds for each variable. It is done by calling the checkWellDesigned() method of
	 * the root node of the tree. If the tree is not well-designed, then a QwdStatus object is stored, containing a counterexample witnessing the violation.
	 * 
	 * If we need a list of all violations, this could also be done, needs only minor changes
	 * 
	 * TODO: note that the variable comparison is based on the getName() method for variables only. This might lead to problems with namespaces?
	 * 
	 * @return
	 */
	public boolean checkWellDesigned() {

		QwdStatus status = new QwdStatus();
		if (this.root == null) {
			return true;
		}

		status = this.root.checkWellDesigned(status, new HashSet<String>());
		if (!status.isQWD()) {
			this.wdCounterExamples = status;
		} else {
			this.wdCounterExamples = null;
		}

		return status.isQWD();
	}

	/**
	 * returns the result based on the last run of CheckWellDesigned() Does NOT take care to recompute the status if the query was changed!!
	 * 
	 * @return true if the last run of checkWellDesigned() did so, false otherwise
	 */
	public boolean isWellDesigned() {
		return this.wdCounterExamples == null;
	}

	public Set<String> getWdCounterexampleVars() {

		if (this.wdCounterExamples == null) {
			return new HashSet<String>();
		}
		return this.wdCounterExamples.getConflicts().keySet();
	}

	public WdConflict getWdConflict(String varname) {
		return this.wdCounterExamples.getConflict(varname);
	}

	/**
	 * checks if the pattern tree is quasi well-designed.
	 * 
	 * if the tree consists of no node, then true is returned. Otherwise it is checked if for every two nodes sharing a variable, there exists a common ancestor of this node also containing this
	 * variable. This is done by calling the checkQuasiWellDesigned() method of the root node of the tree. If the tree is not quasi well-designed, then a QwdStatus object is stored, containing a
	 * counterexample witnessing the violation.
	 * 
	 * If we need a list of all violations, this could also be done, needs only minor changes
	 * 
	 * TODO: note that the variable comparison is based on the getName() method for variables only. This might lead to problems with namespaces?
	 * 
	 * @return
	 */
	public boolean checkQuasiWellDesigned() {

		QwdStatus status = new QwdStatus();
		if (this.root == null) {
			return true;
		}

		status = this.root.checkQuasiWellDesigned(status, new HashSet<String>());
		if (status.isQWD()) {
			this.qwdCounterExamples = null;
			return true;
		} else {
			this.qwdCounterExamples = status;
			return false;
		}
	}

	/**
	 * returns the result based on the last run of CheckQuasiWellDesigned() Does NOT take care to recompute the status if the query was changed!!
	 * 
	 * @return true if the last run of checkQuasiWellDesigned() did so, false otherwise
	 */
	public boolean isQuasiWellDesigned() {
		return this.qwdCounterExamples == null;
	}

	public Set<String> getQWdCounterexampleVars() {

		if (this.qwdCounterExamples == null) {
			return new HashSet<String>();
		}
		return this.qwdCounterExamples.getConflicts().keySet();
	}

	public WdConflict getQwdConflict(String varname) {
		return this.qwdCounterExamples.getConflict(varname);
	}

	public void unsetRuntimeStatistics() {
		this.root.unsetRuntimeStatistics();
	}

	public class QwdStatus {

		private final HashMap<String, TreePatternNode> firstNodes;
		private final HashMap<String, WdConflict> conflicts;

		public QwdStatus() {
			this.firstNodes = new HashMap<String, TreePatternNode>();
			this.conflicts = new HashMap<String, WdConflict>();
		}

		public boolean isQWD() {
			return this.conflicts.isEmpty();
		}

		public HashMap<String, TreePatternNode> getMap() {
			return this.firstNodes;
		}

		public void updateLastNodes(Set<String> variables, TreePatternNode node) {
			for (String var : variables) {
				if (this.firstNodes.containsKey(var)) {
					this.firstNodes.put(var, node);
				}
			}
		}

		public HashMap<String, WdConflict> getConflicts() {
			return this.conflicts;
		}

		public WdConflict getConflict(String key) {
			return this.conflicts.get(key);
		}

		public void addConflict(String var, TreePatternNode n) {

			if (this.conflicts.containsKey(var)) {
				this.conflicts.get(var).addConflictNode(n);

			} else {
				WdConflict c = new WdConflict(var);
				c.addConflictNode(n);
				this.conflicts.put(var, c);
			}
		}
	}

	public class WdConflict {

		private final String variableName;
		private final List<TreePatternNode> nodes;

		public WdConflict(String varname) {
			this.variableName = varname;
			this.nodes = new LinkedList<TreePatternNode>();
		}

		public void addConflictNode(TreePatternNode node) {
			this.nodes.add(node);
		}

		public boolean isConflictNode(TreePatternNode node) {
			return this.nodes.contains(node);
		}

		public String getVarname() {
			return this.variableName;
		}
	}
}
