package spam.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import spam.exception.PatternTreeFactoryException;
import spam.exception.UnsupportedOperatorForPatternTreeException;
import spam.service.OptNormalFormService;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.sparql.algebra.Algebra;
import com.hp.hpl.jena.sparql.algebra.Op;
import com.hp.hpl.jena.sparql.algebra.op.Op1;
import com.hp.hpl.jena.sparql.algebra.op.OpBGP;
import com.hp.hpl.jena.sparql.algebra.op.OpFilter;
import com.hp.hpl.jena.sparql.algebra.op.OpJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpLeftJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpUnion;

/**
 * Creates TreePattern objects from queries or ops
 * 
 * @author Andres Letelier
 */
public class TreePatternFactory {

	/**
	 * In case you're even lazier and don't even want to create the query object from the string
	 * 
	 * @param s
	 * @return
	 * @throws UnsupportedOperatorForPatternTreeException
	 * @throws PatternTreeFactoryException
	 */
	public static TreePattern create(String s) throws PatternTreeFactoryException {
		Query q = QueryFactory.create(s);
		return create(q);
	}

	/**
	 * In case you're too lazy to compile the query yourself
	 * 
	 * @param q
	 * @return
	 * @throws PatternTreeFactoryException
	 * @throws UnsupportedOperatorForPatternTreeException
	 */
	public static TreePattern create(Query q) throws PatternTreeFactoryException {
		if (q == null) {
			return null;
		}
		Op op = Algebra.compile(q);
		try {
			return TreePatternFactory.create(op);
		} catch (UnsupportedOperatorForPatternTreeException e) {
			throw new PatternTreeFactoryException(e);
		}
	}

	/**
	 * Create a well-designed tree pattern from a well-designed query pattern
	 * 
	 * @param op
	 * @return
	 * @throws UnsupportedOperatorForPatternTreeException
	 * @throws PatternTreeFactoryException
	 */
	public static TreePattern create(Op op) throws UnsupportedOperatorForPatternTreeException, PatternTreeFactoryException {

		TreePattern result = new TreePattern();

		if (!OptNormalFormService.isOptNormalForm(op)) {
			result.setStatusString("The query is not in Opt normal form");
		}

		result.setHead(op);
		TreePatternNode rootNode = new TreePatternNode();
		result.setRoot(buildTree(getPattern(op), rootNode));

		return result;
	}

	/**
	 * Build a well-designed tree pattern node from a well-designed query pattern
	 * 
	 * @param op
	 *            the well designed query
	 * @param curNode
	 *            the current position in the tree to be built
	 * @return the tree
	 * @throws UnsupportedOperatorForPatternTreeException
	 * @throws PatternTreeFactoryException
	 */
	protected static TreePatternNode buildTree(Op op, TreePatternNode curNode) throws UnsupportedOperatorForPatternTreeException, PatternTreeFactoryException {
		if (op instanceof OpBGP) {
			curNode.addBP(((OpBGP) op).getPattern());
			return curNode;
		}
		// join operation
		if (op instanceof OpJoin) {
			OpJoin join = (OpJoin) op;
			buildTree(join.getLeft(), curNode);
			buildTree(join.getRight(), curNode);
			return curNode;
		}
		// optional operation
		if (op instanceof OpLeftJoin) {
			OpLeftJoin lj = (OpLeftJoin) op;
			curNode = buildTree(lj.getLeft(), curNode);
			TreePatternNode child;
			if (lj.getExprs() != null) {
				child = buildTree(OpFilter.filterDirect(lj.getExprs(), lj.getRight()), new TreePatternNode());
			} else {
				child = buildTree(lj.getRight(), new TreePatternNode());
			}
			curNode.addChild(child);
			return curNode;
		}
		// filter operation
		if (op instanceof OpFilter) {
			OpFilter f = (OpFilter) op;
			if (f.getSubOp() instanceof OpLeftJoin) {
				throw new PatternTreeFactoryException();
			}
			curNode = buildTree(f.getSubOp(), curNode);
			curNode.addFilter(f.getExprs());
			return curNode;
		}
		if (op instanceof OpUnion) {
			OpUnion u = (OpUnion) op;
			TreePatternNode leftChild = buildTree(u.getLeft(), new TreePatternNode());
			TreePatternNode rightChild = buildTree(u.getRight(), new TreePatternNode());
			curNode.addChild(leftChild);
			curNode.addChild(rightChild);
			// cleanup nested unions
			pullUpUnionNodes(curNode);
			return curNode;
		}
		throw new UnsupportedOperatorForPatternTreeException(op.getClass().getSimpleName());
	}

	/**
	 * pulls up possible nested union nodes into one single union node with more than 2 childs
	 * 
	 * @param curNode
	 *            a union node
	 */
	private static void pullUpUnionNodes(TreePatternNode curNode) {
		boolean again = true;
		while (again) {
			again = false;
			for (TreePatternNode child : new UnionChildIterator(curNode)) {
				for (TreePatternNode childOfUnion : child.getChildren()) {
					if (childOfUnion.getBgp() == null) {
						again = true;
					}
					curNode.addChild(childOfUnion);
				}
				curNode.getChildren().remove(child);
			}
		}
	}

	/**
	 * Fall through useless operations (like projections, orders and the like) until we get to the actual pattern
	 * 
	 * @param op
	 * @throws UnsupportedOperatorForPatternTreeException
	 */
	private static Op getPattern(Op op) throws UnsupportedOperatorForPatternTreeException {
		if (TreePattern.isAcceptedHeadOp(op)) {
			Op1 op1 = (Op1) op;

			return getPattern(op1.getSubOp());
		} else if (TreePattern.isAcceptedBodyOp(op)) {
			return op;
		} else if (op == null) {
			return null;
		}
		throw new UnsupportedOperatorForPatternTreeException(op.getClass().getSimpleName());
	}

	private static class UnionChildIterator implements Iterable<TreePatternNode> {
		static List<TreePatternNode> unionChilds;

		public UnionChildIterator(TreePatternNode unionNode) {
			UnionChildIterator.unionChilds = new ArrayList<TreePatternNode>();
			for (TreePatternNode child : unionNode.getChildren()) {
				if (child.getBgp() == null) {
					unionChilds.add(child);
				}
			}
		}

		@Override
		public Iterator<TreePatternNode> iterator() {
			return unionChilds.iterator();
		}

	}
}
