package spam.model;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import spam.model.TreePattern.QwdStatus;

import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.shared.PrefixMapping;
import com.hp.hpl.jena.sparql.algebra.Algebra;
import com.hp.hpl.jena.sparql.algebra.Op;
import com.hp.hpl.jena.sparql.algebra.OpVars;
import com.hp.hpl.jena.sparql.algebra.op.OpBGP;
import com.hp.hpl.jena.sparql.algebra.op.OpFilter;
import com.hp.hpl.jena.sparql.algebra.op.OpLeftJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpUnion;
import com.hp.hpl.jena.sparql.core.BasicPattern;
import com.hp.hpl.jena.sparql.core.Var;
import com.hp.hpl.jena.sparql.expr.Expr;
import com.hp.hpl.jena.sparql.expr.ExprList;

public class TreePatternNode {

	private ArrayList<TreePatternNode> children;
	private TreePatternNode parent;
	private OpBGP bgp;
	private ExprList filter;
	private long evaluationTime = -1;
	private int numberOfResults = -1;
	private int numberOfLocalMappings = -1;
	private final List<Triple> redundantTriples;

	public TreePatternNode() {
		this.bgp = null;
		this.children = new ArrayList<TreePatternNode>();
		this.parent = null;
		this.redundantTriples = new LinkedList<Triple>();
	}

	public TreePatternNode(OpBGP bgp) {
		this.parent = null;
		this.bgp = bgp;
		this.children = new ArrayList<TreePatternNode>();
		this.redundantTriples = new LinkedList<Triple>();
	}

	public TreePatternNode(String bgp) {

		this.children = new ArrayList<TreePatternNode>();
		this.parent = null;
		this.redundantTriples = new LinkedList<Triple>();

		Query q = QueryFactory.create("SELECT * WHERE { " + bgp + " }");

		this.bgp = (OpBGP) Algebra.compile(q);
	}

	public long getEvaluationTime() {
		return evaluationTime;
	}

	public void setEvaluationTime(long evaluationTime) {
		this.evaluationTime = evaluationTime;
	}

	public int getNumberOfResults() {
		return numberOfResults;
	}

	public void setNumberOfResults(int results) {
		this.numberOfResults = results;
	}

	public int getNumberOfLocalMappings() {
		return numberOfLocalMappings;
	}

	public void setNumberOfLocalMappings(int results) {
		this.numberOfLocalMappings = results;
	}

	public boolean hasChildren() {
		return this.children != null && this.children.size() > 0;
	}

	/**
	 * Invalidate the runtime statistic values.
	 * 
	 * sets them both to -1 in order to indicate that there does not exist valid runtime statistics.
	 */
	public void unsetRuntimeStatistics() {
		this.setEvaluationTime(-1);
		this.setNumberOfResults(-1);

		// And propagate this to the children
		for (TreePatternNode c : this.children) {
			c.unsetRuntimeStatistics();
		}
	}

	/**
	 * @return the root of the tree to which this node belongs
	 */
	public TreePatternNode getRoot() {
		return this.parent != null ? parent.getRoot() : this;
	}

	/*
	 * @return the operator form of the node
	 */
	public Op toOp() {
		Op op = null;
		if (bgp != null) {
			op = bgp;
			if (filter != null && !filter.isEmpty()) {
				op = OpFilter.filter(filter, op);
			}
			for (TreePatternNode child : children) {
				op = OpLeftJoin.create(op, child.toOp(), new ExprList());
			}
		} else {
			// a union node can have more than two child
			op = buildBalancedUnionOpTree(children);
		}
		return op;
	}

	private Op buildBalancedUnionOpTree(ArrayList<TreePatternNode> children) {
		List<Op> opNodes = new ArrayList<>();
		// creating list of (union free) leafs
		for (TreePatternNode tpNode : children) {
			opNodes.add(tpNode.toOp());
		}
		// joining them till only one root element is left
		while (opNodes.size() > 1) {
			List<Op> newOpNodes = new ArrayList<>();
			if ((opNodes.size() % 2) != 0) {
				// remove last element if odd number of nodes
				newOpNodes.add(opNodes.get(opNodes.size() - 1));
				opNodes.remove(opNodes.size() - 1);
			}
			for (int j = 0; j < opNodes.size() - 1; j = j + 2) {
				Op newLeft = opNodes.get(j);
				Op newRight = opNodes.get(j + 1);
				newOpNodes.add(OpUnion.create(newLeft, newRight));
			}
			opNodes.clear();
			opNodes.addAll(newOpNodes);
		}
		return opNodes.get(0);
	}

	/**
	 * Add a bunch of filter expressions, checking that they only mention variables that appear in the pattern
	 * 
	 * @param other
	 */
	public void mergeFilter(ExprList other) {
		Set<Var> vars = new HashSet<Var>(OpVars.mentionedVars(bgp));
		exprs: for (Expr ex : other) {
			Set<Var> otherVars = ex.getVarsMentioned();
			for (Var x : otherVars) {
				if (!vars.contains(x)) {
					continue exprs;
				}
			}
		}
	}

	/**
	 * Add a bunch of filter expressions without checking them
	 * 
	 * @param other
	 */
	public void addFilter(ExprList other) {
		if (filter == null) {
			filter = new ExprList();
			filter.addAll(other);
		} else {
			filter.addAll(other);
		}
	}

	/**
	 * Add one filter expression without checking it
	 * 
	 * @param exp
	 */
	public void addFilter(Expr exp) {
		if (filter == null) {
			filter = new ExprList(exp);
		} else {
			filter.add(exp);
		}
	}

	public ExprList getFilter() {
		return filter;
	}

	public void setFilter(ExprList filter) {
		if (filter == null) {
			return;
		}
		if (this.filter == null) {
			this.filter = filter;
		} else {
			this.filter.addAll(filter);
		}
	}

	public void addChild(TreePatternNode child) {
		children.add(child);
		child.parent = this;
	}

	public OpBGP getBgp() {
		return bgp;
	}

	public void setBgp(OpBGP bgp) {
		this.bgp = bgp;
	}

	public boolean setBgp(String bgp, PrefixMapping prefixes) throws Exception {
		// TODO: it must be possible to compile a given BGP without
		// all this trouble
		// Op newOp = Algebra.parse(newContent);
		// BasicPattern bgp = SSE.parseBGP(newContent);
		// But none of the above works

		StringBuilder prefixString = new StringBuilder();
		if (prefixes != null && prefixes.toString().length() > 5) {
			String[] ps = prefixes.toString().substring(4, prefixes.toString().length() - 1).split(",");
			for (int i = 0; i < ps.length; i++) {
				String[] m = ps[i].split("=");
				prefixString.append("PREFIX ");
				prefixString.append(m[0]);
				prefixString.append(": <");
				prefixString.append(m[1]);
				prefixString.append(">\n");
			}
		}

		Query q = QueryFactory.create(prefixString + " SELECT * WHERE { " + bgp + " }");

		q.setPrefixMapping(prefixes);
		Op op = Algebra.compile(q);
		if (!(op instanceof OpBGP)) {
			return false;
		}

		this.bgp = (OpBGP) op;
		return true;

	}

	public void addBP(BasicPattern newPattern) {

		BasicPattern oldPattern;
		if (this.bgp != null) {
			oldPattern = this.bgp.getPattern();
		} else {
			oldPattern = new BasicPattern();
		}
		oldPattern.addAll(newPattern);
		this.setBgp(new OpBGP(oldPattern));
	}

	public TreePatternNode getParent() {
		return parent;
	}

	public void setParent(TreePatternNode parent) {
		this.parent = parent;
	}

	/**
	 * Merge the node into its parent.
	 * 
	 * Merging means adding every child of this node as children its parent and adding this nodes' triples as triples of the parent. Finally, delete this node from the parent's list of children
	 */
	public void mergeIntoParent() {
		parent.bgp.getPattern().addAll(this.bgp.getPattern());
		parent.children.addAll(this.children);
		for (TreePatternNode child : children) {
			child.setParent(parent);
		}
		this.parent.children.remove(this);
	}

	/**
	 * Merge a node into its children. This means adding this node's pattern into each of its children, making each child a child of this node's parent and deleting this node.
	 */
	public void mergeIntoChildren() {
		for (TreePatternNode child : children) {
			child.bgp.getPattern().addAll(this.bgp.getPattern());
		}
		this.delete();
	}

	/**
	 * Deletes the current node, making every child a son of this nodes' parent.
	 */
	public void delete() {
		if (this.isRoot()) {
			System.err.println("Tried to delete root node!");
			return;
		}
		for (TreePatternNode child : children) {
			child.parent = this.parent;
		}

		parent.children.remove(this);
		parent.children.addAll(children);
	}

	public void deleteSubtree() {

		if (parent != null) {
			parent.children.remove(this);
		} else {
			this.children.clear();
		}
	}

	public boolean moveNode(TreePatternNode newParent) {

		if (this.parent == null) {
			// root cannot be moved
			return false;
		}

		this.parent.children.remove(this);
		this.parent = newParent;
		if (this.parent != null) {
			this.parent.addChild(this);
		}

		return true;
	}

	/**
	 * Returns a unique string for this object (hopefully)
	 * 
	 * @return
	 */
	public String getId() {
		return super.toString();

	}

	@Override
	public String toString() {
		if (bgp == null) {
			return "UNION";
		} else {
			return bgp.getPattern().toString();
		}

	}

	public ArrayList<TreePatternNode> getChildren() {
		return children;
	}

	public boolean isRoot() {
		return this.parent == null;
	}

	/**
	 * dump into console for debugging
	 */
	public void dump() {
		dump("");
	}

	/**
	 * dump into console for debugging
	 * 
	 * @param prec
	 */
	private void dump(String prec) {
		System.err.println(prec + bgp);
		for (TreePatternNode child : children) {
			child.dump(prec + "  ");
		}
	}

	public void addRedundantTriple(Triple t) {
		this.redundantTriples.add(t);
	}

	/**
	 * returns the (unqualified !!!) names of the variables used in the BGP at this node
	 * 
	 * @return the set of unqualified names
	 */
	public Set<String> getVariableNames() {
		Set<String> names = new HashSet<String>();
		if (this.bgp == null) {
			return names;
		}

		for (Triple t : this.bgp.getPattern()) {
			if (t.getSubject().isVariable()) {
				names.add(t.getSubject().getName());
			}
			if (t.getPredicate().isVariable()) {
				names.add(t.getPredicate().getName());
			}
			if (t.getObject().isVariable()) {
				names.add(t.getObject().getName());
			}
		}
		return names;
	}

	public Set<String> getVariableNamesIncludingFilter() {
		Set<String> vars = this.getVariableNames();
		if (this.getFilter() != null) {
			for (Var var : this.getFilter().getVarsMentioned()) {
				vars.add(var.getName());
			}
		}
		return vars;
	}

	public QwdStatus checkQuasiWellDesigned(QwdStatus status, Set<String> oldVariables) {

		Set<String> curVars = this.getVariableNames();
		Set<String> newVars = new HashSet<String>();
		Set<String> visitedVars = new HashSet<String>();

		newVars.addAll(curVars);
		newVars.removeAll(oldVariables);

		for (String v : newVars) {
			if (status.getMap().containsKey(v)) {

				// The pattern ist NOT qwd
				status.addConflict(v, status.getMap().get(v));
				status.addConflict(v, this);

			} else {

				// This variable does not ruin qwd
				// hence we have to mark this node as introducing v
				status.getMap().put(v, this);
			}
		}

		if (this.filter != null) {
			for (Var var : filter.getVarsMentioned()) {
				// (P FILTER R) -> vars(R) subset of vars(P)
				if (!curVars.contains(var.getName())) {
					status.addConflict(var.getName(), this);
				}
			}
		}

		visitedVars.addAll(oldVariables);
		visitedVars.addAll(curVars);

		// iterate into the subtrees
		for (TreePatternNode n : this.children) {
			if (this.bgp == null) {
				// UNION NODE
				if (status.isQWD()) {
					status.getMap().clear();
				}
			}
			status = n.checkQuasiWellDesigned(status, visitedVars);
		}
		return status;
	}

	/**
	 * checks if the nodes violates the condition of being well-designed.
	 * 
	 * Using the information of a global map that stores at which node a variables was seen for the first time, all variables at this node are compared with the variables at the parent node. If for
	 * some new variable (i.e. for a local variable that does not occur in the parent node), there already exists an entry in the map, then there is a violation of the connectedness condition (since
	 * we are doing a DFS traversal
	 * 
	 * @param status
	 * @param oldVariables
	 * @return
	 */
	public QwdStatus checkWellDesigned(QwdStatus status, Set<String> parentVariables) {

		Set<String> curVars = this.getVariableNames();
		Set<String> newVars = new HashSet<String>();

		newVars.addAll(curVars);
		newVars.removeAll(parentVariables);

		for (String v : newVars) {
			if (status.getMap().containsKey(v)) {
				// The pattern ist NOT wd
				status.addConflict(v, status.getMap().get(v));
				status.addConflict(v, this);
			} else {
				// This variable does not ruin wd
				// hence we have to mark this node as introducing v
				status.getMap().put(v, this);
			}
		}

		if (this.filter != null) {
			for (Var var : filter.getVarsMentioned()) {
				// (P FILTER R) -> vars(R) subset of vars(P)
				if (!curVars.contains(var.getName())) {
					status.addConflict(var.getName(), this);
				}
			}
		}

		// iterate into the subtrees
		for (TreePatternNode n : this.children) {
			if (this.bgp == null) {
				// UNION NODE
				if (status.isQWD()) {
					status.getMap().clear();
				}
			}
			status = n.checkWellDesigned(status, curVars);
		}
		return status;

	}

	/**
	 * checks if the specified BGP is equal to the set of triple patterns stored at this node.
	 * 
	 * @param bgp
	 *            the BGP to compare with the local set of triple patterns
	 * @return true if they are the same, false otherwise
	 */
	public boolean equalsBGP(OpBGP bgp) {

		if (this.bgp == null) {
			return false;
		} else {
			return this.bgp.equalTo(bgp, null);
		}
	}

	public boolean equalsBGP(TreePatternNode pattern) {
		return this.equalsBGP(pattern.getBgp());
	}

	public StringBuffer getDebugString(int level) {
		StringBuffer o = new StringBuffer();
		o = addIndent(o, level);
		o.append("Node content: ");

		int c = 0;
		for (Triple t : this.bgp.getPattern()) {
			if (c > 0) {
				o.append(" . ");
			}
			o.append(t.toString());
			c++;
		}
		o.append("\n");
		o = addIndent(o, level);
		o.append("Children: [");
		for (TreePatternNode n : children) {
			o.append("\n");
			o = addIndent(o, level + 1);
			o.append(n.getDebugString(level + 1));
		}
		if (children.size() > 0) {
			o.append("\n");
			o = addIndent(o, level);
		}
		o.append("]");
		return o;
	}

	private static StringBuffer addIndent(StringBuffer o, int level) {
		for (int i = 0; i < level; i++) {
			o.append(" ");
		}
		return o;
	}

	/**
	 * Deep copy of this node
	 * 
	 * @return
	 */
	protected TreePatternNode deepCopy() {
		TreePatternNode copy = new TreePatternNode();
		if (this.bgp != null) {
			copy.bgp = new OpBGP(new BasicPattern(this.bgp.getPattern()));
		}
		copy.evaluationTime = this.evaluationTime;
		copy.numberOfResults = this.numberOfResults;
		if (filter != null) {
			copy.filter = new ExprList(this.filter.getList());
		}
		copy.children = new ArrayList<TreePatternNode>();

		for (TreePatternNode child : this.children) {
			TreePatternNode childcopy = child.deepCopy();
			childcopy.parent = copy;
			copy.children.add(childcopy);
		}
		return copy;
	}

	public boolean isRedundantTriple(Triple t) {
		return this.redundantTriples.contains(t);
	}

	public boolean isAncestor(TreePatternNode node) {

		if (this.parent == null) {
			return false;
		}

		if (this.parent == node) {
			return true;
		}

		return this.parent.isAncestor(node);
	}

	public List<TreePatternNode> getPathToAncestor(TreePatternNode node, List<TreePatternNode> prefix) {

		if (this.parent == null) {
			return null;
		}

		prefix.add(this);

		if (this.parent == node) {
			return prefix;
		}

		return this.parent.getPathToAncestor(node, prefix);
	}
}
