package spam.service;

import spam.exception.UntransformableOperatorException;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.shared.PrefixMapping;
import com.hp.hpl.jena.sparql.algebra.Algebra;
import com.hp.hpl.jena.sparql.algebra.Op;
import com.hp.hpl.jena.sparql.algebra.OpAsQuery;
import com.hp.hpl.jena.sparql.algebra.op.Op0;
import com.hp.hpl.jena.sparql.algebra.op.Op1;
import com.hp.hpl.jena.sparql.algebra.op.OpBGP;
import com.hp.hpl.jena.sparql.algebra.op.OpDistinct;
import com.hp.hpl.jena.sparql.algebra.op.OpFilter;
import com.hp.hpl.jena.sparql.algebra.op.OpJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpLeftJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpOrder;
import com.hp.hpl.jena.sparql.algebra.op.OpProject;
import com.hp.hpl.jena.sparql.algebra.op.OpReduced;
import com.hp.hpl.jena.sparql.algebra.op.OpService;
import com.hp.hpl.jena.sparql.algebra.op.OpSlice;
import com.hp.hpl.jena.sparql.algebra.op.OpUnion;

public class OptNormalFormService {

	/**
	 * Checks if the query is in OPT-normal form as defined in "Semantics and Complexity of SPARQL" by Perez et al, 2009.
	 * 
	 * @param q
	 *            The query to be tested
	 * @return true if the query is in OPT-normal form
	 */
	public static boolean isOptNormalForm(Query q) {
		Op op = Algebra.compile(q);
		return isOptNormalForm(op);
	}

	/**
	 * Checks if the query is in OPT-normal form as defined in "Semantics and Complexity of SPARQL" by Perez et al, 2009.
	 * 
	 * @param op
	 *            the pattern to be tested
	 * @return true if the pattern is in OPT-normal form
	 */
	public static boolean isOptNormalForm(Op op) {
		return isOptNormalFormWorker(op, false, true);
	}

	/**
	 * checks recursively if the given pattern is in OPT-normal form as defined in "Semantics and Complexity of SPARQL" by Perez et al, 2009.
	 * 
	 * The boolean parameter tells if nested optionals are allowed at this point, or if the current operator must basically be a CQ (in one of many possible syntactic representations)
	 * 
	 * @param op
	 *            The operator to check for being in OPT-normal form
	 * @param basicFormOnly
	 *            true if the pattern must be equivalent to a BGP
	 * @param unionAllowed
	 *            true if union is allowed inside this operator
	 * @return true if the pattern is in OPT normal form, relative to "basicFormOnly"
	 */
	private static boolean isOptNormalFormWorker(Op op, boolean basicFormOnly, boolean unionAllowed) {
		if (op instanceof OpProject || op instanceof OpSlice || op instanceof OpReduced || op instanceof OpOrder || op instanceof OpDistinct || op instanceof OpService || op instanceof OpFilter) {
			Op sub = ((Op1) op).getSubOp();
			return isOptNormalFormWorker(sub, basicFormOnly, true);
		}
		if (op instanceof OpUnion) {
			if (unionAllowed == false) {
				// union not in top level detected
				return false;
			}
			OpUnion union = (OpUnion) op;
			return isOptNormalFormWorker(union.getLeft(), basicFormOnly, true) && isOptNormalFormWorker(union.getRight(), basicFormOnly, true);
		}

		if (op instanceof OpBGP) {
			return true;
		}
		if (op instanceof OpJoin) {
			OpJoin join = (OpJoin) op;
			return isOptNormalFormWorker(join.getLeft(), true, false) && isOptNormalFormWorker(join.getRight(), true, false);
		}
		if (op instanceof OpLeftJoin) {
			OpLeftJoin lj = (OpLeftJoin) op;
			return !basicFormOnly && isOptNormalFormWorker(lj.getLeft(), false, false) && isOptNormalFormWorker(lj.getRight(), false, false);
		}

		if (op instanceof Op0) {
			return true;
		}
		if (op == null) {
			return true;
		}
		throw new UnsupportedOperationException(op.toString());
	}

	public static Query getOptNormalForm(final Query original) throws UntransformableOperatorException {
		Op root = Algebra.compile(original);

		root = convertertToOptNormalForm(root, true);

		PrefixMapping pm = original.getPrefixMapping();
		Query newQ = OpAsQuery.asQuery(root);
		newQ.setPrefixMapping(pm);
		return newQ;
	}

	/**
	 * Converts an Operator Tree into Opt Normal Form if possible, taking UNION and FILTER into account All Union operators are pulled into the top level
	 * 
	 * @param root
	 *            the root of the Op Tree that will be converted
	 * @param otherOps
	 *            are non Pattern Matching Operators allowed at this place?
	 * @throws UntransformableOperatorException
	 */
	private static Op convertertToOptNormalForm(Op op, boolean otherOPs) throws UntransformableOperatorException {
		if (op instanceof OpBGP) {
			return op;
		}
		if (op instanceof OpUnion) {
			OpUnion union = (OpUnion) op;
			Op left = convertertToOptNormalForm(union.getLeft(), false);
			Op right = convertertToOptNormalForm(union.getRight(), false);
			return OpUnion.create(left, right);
		}
		if (op instanceof OpJoin) {
			OpJoin join = (OpJoin) op;
			if (join.getLeft() instanceof OpUnion) {
				// (P1 UNION P2) AND P3
				return distributeJoinOverUnion((OpUnion) join.getLeft(), join.getRight());
			} else if (join.getRight() instanceof OpUnion) {
				// P3 AND (P1 UNION P2)
				return distributeJoinOverUnion((OpUnion) join.getRight(), join.getLeft());
			} else {
				// continue with childs of join
				Op left = convertertToOptNormalForm(join.getLeft(), false);
				Op right = convertertToOptNormalForm(join.getRight(), false);
				return OpJoin.create(left, right);
			}
		}
		if (op instanceof OpLeftJoin) {
			OpLeftJoin opt = (OpLeftJoin) op;
			if (opt.getLeft() instanceof OpUnion) {
				// (P1 UNION P2) OPT P3
				Op sub = opt.getRight();
				if (opt.getExprs() != null) {
					sub = OpFilter.filterDirect(opt.getExprs(), sub);
				}
				Op left = OpLeftJoin.createLeftJoin(((OpUnion) opt.getLeft()).getLeft(), opt.getRight(), null);
				Op right = OpLeftJoin.createLeftJoin(((OpUnion) opt.getLeft()).getRight(), opt.getRight(), null);
				Op newLeft = convertertToOptNormalForm(left, false);
				Op newRight = convertertToOptNormalForm(right, false);
				return new OpUnion(newLeft, newRight);
			} else if (opt.getRight() instanceof OpUnion) {
				// P3 OPT (P1 UNION P2)
				Op left = OpLeftJoin.createLeftJoin(opt.getLeft(), ((OpUnion) opt.getRight()).getLeft(), null);
				Op right = OpLeftJoin.createLeftJoin(opt.getLeft(), ((OpUnion) opt.getRight()).getRight(), null);
				if (opt.getExprs() != null) {
					left = OpFilter.filterDirect(opt.getExprs(), left);
					right = OpFilter.filterDirect(opt.getExprs(), right);
				}
				Op newLeft = convertertToOptNormalForm(left, false);
				Op newRight = convertertToOptNormalForm(right, false);
				return new OpUnion(newLeft, newRight);
			} else {
				// continue with childs of join
				Op left = convertertToOptNormalForm(opt.getLeft(), false);
				Op right = convertertToOptNormalForm(opt.getRight(), false);
				if (opt.getExprs() != null) {
					right = OpFilter.filterDirect(opt.getExprs(), right);
					right = convertertToOptNormalForm(right, false);
				}
				Op newOp = OpLeftJoin.createLeftJoin(left, right, null);
				return newOp;
			}
		}
		if (op instanceof OpFilter) {
			OpFilter filter = (OpFilter) op;
			if (filter.getSubOp() instanceof OpUnion) {
				// ((P1 UNION P2) FILTER R) -> ((P1 FILTER R) UNION (P2 FILTER R))
				OpUnion union = (OpUnion) filter.getSubOp();
				Op newLeft = OpFilter.filterDirect(filter.getExprs(), union.getLeft());
				Op newRight = OpFilter.filterDirect(filter.getExprs(), union.getRight());
				newLeft = convertertToOptNormalForm(newLeft, false);
				newRight = convertertToOptNormalForm(newRight, false);
				return OpUnion.create(newLeft, newRight);
			} else {
				return OpFilter.filter(filter.getExprs(), convertertToOptNormalForm(filter.getSubOp(), false));
			}
		}
		// now come the other operators
		if (!otherOPs) {
			throw new UntransformableOperatorException();
		}
		if (op instanceof OpProject) {
			OpProject opProject = (OpProject) op;
			return new OpProject(convertertToOptNormalForm(opProject.getSubOp(), true), opProject.getVars());
		}
		if (op instanceof OpSlice) {
			OpSlice opSlice = (OpSlice) op;
			return new OpSlice(convertertToOptNormalForm(opSlice.getSubOp(), true), opSlice.getStart(), opSlice.getLength());
		}
		if (op instanceof OpReduced) {
			OpReduced opReduced = (OpReduced) op;
			return OpReduced.create(convertertToOptNormalForm(opReduced.getSubOp(), true));
		}
		if (op instanceof OpOrder) {
			OpOrder opOrder = (OpOrder) op;
			return new OpOrder(convertertToOptNormalForm(opOrder.getSubOp(), true), opOrder.getConditions());
		}
		if (op instanceof OpDistinct) {
			OpDistinct opDistinct = (OpDistinct) op;
			return OpDistinct.create(convertertToOptNormalForm(opDistinct.getSubOp(), true));
		}
		throw new UntransformableOperatorException();
	}

	/**
	 * Transforms (P1 UNION P2) AND P3 ---> (P1 AND P3) UNION (P2 AND P3)
	 * 
	 * @param p1
	 *            left side of sub union
	 * @param p2
	 *            right side of sub union
	 * @param p3
	 *            right side of original operator
	 * @return the resulting union op
	 * @throws UntransformableOperatorException
	 */
	private static OpUnion distributeJoinOverUnion(OpUnion union, Op p3) throws UntransformableOperatorException {
		OpJoin left = (OpJoin) OpJoin.create(union.getLeft(), p3);
		OpJoin right = (OpJoin) OpJoin.create(union.getRight(), p3);

		Op newLeft = convertertToOptNormalForm(left, false);
		Op newRight = convertertToOptNormalForm(right, false);
		return new OpUnion(newLeft, newRight);
	}

	/**
	 * Converts a well designed query, so that it can be converted to a valid pattern tree later on.
	 * 
	 * Every occuring of the Pattern (P1 OPT P2) FILTER R -> (P1 FILTER R) OPT P2 as defined in "Semantics and Complexity of SPARQL" by Perez et al, 2009.
	 * 
	 * @param original
	 *            a well designed query
	 * @return a new Query that can be safely converted to a Pattern Tree
	 * @throws UntransformableOperatorException
	 *             if such operations occur
	 */
	public static Query getTreePatternConvertableForm(Query original) throws UntransformableOperatorException {
		Op root = Algebra.compile(original);

		root = convertToTreePatternConvertableForm(root, true);

		PrefixMapping pm = original.getPrefixMapping();
		Query newQ = OpAsQuery.asQuery(root);
		newQ.setPrefixMapping(pm);
		return newQ;
	}

	private static Op convertToTreePatternConvertableForm(Op op, boolean otherOPs) throws UntransformableOperatorException {
		if (op instanceof OpBGP) {
			return op;
		}
		if (op instanceof OpUnion) {
			OpUnion union = (OpUnion) op;
			Op left = convertToTreePatternConvertableForm(union.getLeft(), false);
			Op right = convertToTreePatternConvertableForm(union.getRight(), false);
			return OpUnion.create(left, right);
		}
		if (op instanceof OpJoin) {
			OpJoin join = (OpJoin) op;
			Op left = convertToTreePatternConvertableForm(join.getLeft(), false);
			Op right = convertToTreePatternConvertableForm(join.getRight(), false);
			return OpJoin.create(left, right);
		}
		if (op instanceof OpLeftJoin) {
			OpLeftJoin opt = (OpLeftJoin) op;
			Op left = convertToTreePatternConvertableForm(opt.getLeft(), false);
			Op right = convertToTreePatternConvertableForm(opt.getRight(), false);
			if (opt.getExprs() != null) {
				right = OpFilter.filterDirect(opt.getExprs(), right);
				right = convertToTreePatternConvertableForm(right, false);
			}
			Op newOp = OpLeftJoin.createLeftJoin(left, right, null);
			return newOp;
		}
		if (op instanceof OpFilter) {
			OpFilter filter = (OpFilter) op;
			if (filter.getSubOp() instanceof OpLeftJoin) {
				// ((P1 OPT P2) FILTER R) -> ((P1 FILTER R) OPT P2)
				OpLeftJoin opt = (OpLeftJoin) filter.getSubOp();
				OpFilter newLeft = (OpFilter) OpFilter.filter(filter.getExprs(), opt.getLeft());
				return convertToTreePatternConvertableForm(OpLeftJoin.createLeftJoin(newLeft, opt.getRight(), null), false);
			} else if (filter.getSubOp() instanceof OpFilter) {
				// pulling multiple Filters together
				OpFilter subFilter = (OpFilter) filter.getSubOp();
				filter.getExprs().addAll(subFilter.getExprs());
				OpFilter newFilter = OpFilter.filterDirect(filter.getExprs(), subFilter.getSubOp());
				return convertToTreePatternConvertableForm(newFilter, false);
			} else {
				return OpFilter.filter(filter.getExprs(), convertToTreePatternConvertableForm(filter.getSubOp(), false));
			}
		}

		// now come the other operators
		if (!otherOPs) {
			throw new UntransformableOperatorException();
		}
		if (op instanceof OpProject) {
			OpProject opProject = (OpProject) op;
			return new OpProject(convertToTreePatternConvertableForm(opProject.getSubOp(), true), opProject.getVars());
		}
		if (op instanceof OpSlice) {
			OpSlice opSlice = (OpSlice) op;
			return new OpSlice(convertToTreePatternConvertableForm(opSlice.getSubOp(), true), opSlice.getStart(), opSlice.getLength());
		}
		if (op instanceof OpReduced) {
			OpReduced opReduced = (OpReduced) op;
			return OpReduced.create(convertToTreePatternConvertableForm(opReduced.getSubOp(), true));
		}
		if (op instanceof OpOrder) {
			OpOrder opOrder = (OpOrder) op;
			return new OpOrder(convertToTreePatternConvertableForm(opOrder.getSubOp(), true), opOrder.getConditions());
		}
		if (op instanceof OpDistinct) {
			OpDistinct opDistinct = (OpDistinct) op;
			return OpDistinct.create(convertToTreePatternConvertableForm(opDistinct.getSubOp(), true));
		}
		throw new UntransformableOperatorException();
	}
}
