package spam.service;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import spam.exception.UnsupportedOperatorForSubsumptionException;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.sparql.algebra.Algebra;
import com.hp.hpl.jena.sparql.algebra.Op;
import com.hp.hpl.jena.sparql.algebra.OpVars;
import com.hp.hpl.jena.sparql.algebra.op.Op1;
import com.hp.hpl.jena.sparql.algebra.op.Op2;
import com.hp.hpl.jena.sparql.algebra.op.OpBGP;
import com.hp.hpl.jena.sparql.algebra.op.OpFilter;
import com.hp.hpl.jena.sparql.algebra.op.OpJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpLeftJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpModifier;
import com.hp.hpl.jena.sparql.algebra.op.OpProject;
import com.hp.hpl.jena.sparql.algebra.op.OpUnion;
import com.hp.hpl.jena.sparql.core.Var;
import com.hp.hpl.jena.sparql.engine.QueryIterator;
import com.hp.hpl.jena.sparql.engine.binding.Binding;
import com.hp.hpl.jena.sparql.graph.GraphFactory;

public class SubsumptionService {

	public static Graph failGraph;
	private static Op failReduction;

	/**
	 * Tests whether A is equivalent to B
	 * 
	 * @return 0 if A is equivalent to B, 1 if A is not subsumed by B and 2 if B is not subsumed by A
	 * @throws UnsupportedOperatorForSubsumptionException
	 */
	public static int isEquivalent(Op A, Op B) throws UnsupportedOperatorForSubsumptionException {
		if (!isSubsumed(A, B)) {
			return 1;
		}

		if (!isSubsumed(B, A)) {
			return 2;
		}
		return 0;
	}

	public static boolean isSubsumed(Op A, Op B) throws UnsupportedOperatorForSubsumptionException {
		// get all union free subtrees
		List<Op> topOpsA = getTopOps(A);
		List<Op> topOpsB = getTopOps(B);

		// check if every subtree in A is subsumed by a subtree in B
		for (Op topOpA : topOpsA) {
			boolean isSubsumed = false;
			for (Op topOpB : topOpsB) {
				if (isUnionFreeSubsumed(topOpA, topOpB)) {
					isSubsumed = true;
					break;
				}
			}
			if (!isSubsumed) {
				// there is a subtree in A that is not subsumed by a subtree in B
				return false;
			}
		}
		// all subtrees of A are subsumed by at least one subtree of B
		return true;
	}

	// returns the first Operator with 0 or 2 operands
	private static List<Op> getTopOps(Op op) {
		List<Op> result = new ArrayList<>();
		while (op instanceof Op1) {
			op = ((Op1) op).getSubOp();
		}
		if (op instanceof OpUnion) {
			return addAllUnionFreeSubTrees((OpUnion) op, result);
		} else {
			result.add(op);
		}
		return result;
	}

	private static List<Op> addAllUnionFreeSubTrees(OpUnion op, List<Op> result) {
		if (op.getLeft() instanceof OpUnion) {
			addAllUnionFreeSubTrees((OpUnion) op.getLeft(), result);
		} else {
			result.add(op.getLeft());
		}
		if (op.getRight() instanceof OpUnion) {
			addAllUnionFreeSubTrees((OpUnion) op.getRight(), result);
		} else {
			result.add(op.getRight());
		}
		return result;
	}

	/**
	 * Tests if A is subsumed in B: A is subsumed in B iif for every reduction A' of A there exists a reduction B' of B s.t.: a) Vars(A') subset Vars(B') b) There exists a homomorphism from Vars(B')
	 * to Vars(A') s.t. the homomorphism is the identity on Vars(A')
	 * 
	 * @param A
	 *            a union free query pattern
	 * @param B
	 *            another union free query pattern
	 * @return true if A is contained in B
	 * @throws UnsupportedOperatorForSubsumptionException
	 */
	private static boolean isUnionFreeSubsumed(Op A, Op B) throws UnsupportedOperatorForSubsumptionException {
		List<Op> redA = getReductions(A);
		List<Op> redB = getReductions(B);
		RedsA: for (Op rA : redA) {
			Graph gA = GraphFactory.createDefaultGraph();
			addTriples(gA, rA);
			Set<Var> varA = new HashSet<Var>(OpVars.mentionedVars(rA));

			for (Op rB : redB) {
				QueryIterator res = Algebra.exec(rB, gA);
				bindings: while (res.hasNext()) {
					Binding b = res.next();
					for (Var v : varA) {
						if (!b.contains(v) || !b.get(v).matches(v)) {
							continue bindings;
						}
					}
					res.close();
					continue RedsA;
				}
			}
			// if we got here, we find that for database gA both patterns return
			// different thing
			failGraph = gA;
			failReduction = rA;
			return false;
		}
		return true;
	}

	/**
	 * @author Sebastian Skritek
	 * 
	 *         example code taken from http://d2rq.org/jena#usingfind
	 */
	public static Iterator<Triple> getFailGraphTriples() {

		if (failGraph == null) {
			return null;
		}

		// Query the graph
		return retrieveAllTriples(failGraph);
	}

	private static Iterator<Triple> retrieveAllTriples(Graph graph) {

		// Create a find(spo) pattern
		Triple pattern = new Triple(Node.ANY, Node.ANY, Node.ANY);

		// Query the graph
		return graph.find(pattern);
	}

	public static Op getFailReduction() {
		return failReduction;
	}

	private static Set<Var> getQueryVars(Query q) {
		Op qop = Algebra.compile(q);
		Op tmp = qop;
		// Strip the solution modifiers
		while (tmp instanceof OpModifier) {
			tmp = ((OpModifier) tmp).getSubOp();
		}
		Set<Var> vars = new HashSet<Var>(OpVars.mentionedVars(tmp));
		return vars;
	}

	public static boolean compareVarsForSubsumption(Query left, Query right) {
		Set<Var> v1 = getQueryVars(left);
		Set<Var> v2 = getQueryVars(right);
		return v2.containsAll(v1);
	}

	public static boolean compareVarsForEquivalence(Query left, Query right) {
		Set<Var> v1 = getQueryVars(left);
		Set<Var> v2 = getQueryVars(right);
		return v2.containsAll(v1) && v1.containsAll(v2);
	}

	/**
	 * Adds every triple in op to g, creating op's frozen database
	 * 
	 * @param g
	 * @param op
	 */
	private static void addTriples(Graph g, Op op) {
		if (op instanceof OpBGP) {
			OpBGP bgp = (OpBGP) op;
			for (Triple t : bgp.getPattern()) {
				g.add(t);
			}
		}
		if (op instanceof Op2) {
			Op2 op2 = (Op2) op;
			addTriples(g, op2.getLeft());
			addTriples(g, op2.getRight());
			return;
		}
		if (op instanceof Op1) {
			Op1 op1 = (Op1) op;
			addTriples(g, op1.getSubOp());
		}
	}

	/**
	 * Returns a list of all the possible reductions of op.
	 * 
	 * @param op
	 *            the pattern for which reductions will be generated
	 * @return a list of all the reductions of op
	 * @throws UnsupportedOperatorForSubsumptionException
	 *             if it encounters Filters or Projection
	 */
	private static ArrayList<Op> getReductions(Op op) throws UnsupportedOperatorForSubsumptionException {
		ArrayList<Op> red = new ArrayList<Op>();
		if (op instanceof OpBGP) {
			red.add(op);
			return red;
		}
		if (op instanceof OpProject) {
			throw new UnsupportedOperatorForSubsumptionException();
		}
		if (op instanceof OpFilter) {
			throw new UnsupportedOperatorForSubsumptionException();
		}
		if (op instanceof OpLeftJoin) {
			OpLeftJoin lj = (OpLeftJoin) op;
			List<Op> left = getReductions(lj.getLeft());
			List<Op> right = getReductions(lj.getRight());
			for (Op l : left) {
				for (Op r : right) {
					Op lj2 = OpLeftJoin.create(l, r, ((OpLeftJoin) op).getExprs());
					red.add(lj2);
				}
			}
			red.add(lj.getLeft());
			return red;
		}
		if (op instanceof OpJoin) {
			OpJoin j = (OpJoin) op;
			List<Op> left = getReductions(j.getLeft());
			List<Op> right = getReductions(j.getRight());
			for (Op l : left) {
				for (Op r : right) {
					Op j2 = OpJoin.create(l, r);
					red.add(j2);
				}
			}
			return red;
		}

		System.err.println("Not supported Operator: " + op.toString());
		return null;
	}

	/**
	 * Get variables in op that are outside subop
	 * 
	 * @param op
	 *            The complete pattern
	 * @param subop
	 *            A subpattern of op
	 * @return Variables in op without subop
	 */
	protected static Set<Var> getOtherVars(Op op, Op subop) {
		if (op == subop) {
			return new HashSet<Var>();
		}

		if (op instanceof OpFilter) {
			OpFilter f = (OpFilter) op;
			Set<Var> vars = getOtherVars(f.getSubOp(), subop);
			vars.addAll(f.getExprs().getVarsMentioned());
			return vars;
		}
		if (op instanceof OpProject) {
			return getOtherVars(((OpProject) op).getSubOp(), subop);
		}
		if (op instanceof Op2) {
			Op2 op2 = (Op2) op;
			Set<Var> left = getOtherVars(op2.getLeft(), subop), right = getOtherVars(op2.getRight(), subop);
			left.addAll(right);
			return left;
		}
		if (op instanceof OpBGP) {
			return new HashSet<Var>(OpVars.mentionedVars(op));
		}
		throw new UnsupportedOperationException(op.toString());
	}
}
