package spam.service;

import java.util.HashSet;
import java.util.Set;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.sparql.algebra.Algebra;
import com.hp.hpl.jena.sparql.algebra.Op;
import com.hp.hpl.jena.sparql.algebra.OpVars;
import com.hp.hpl.jena.sparql.algebra.op.Op1;
import com.hp.hpl.jena.sparql.algebra.op.OpBGP;
import com.hp.hpl.jena.sparql.algebra.op.OpDistinct;
import com.hp.hpl.jena.sparql.algebra.op.OpFilter;
import com.hp.hpl.jena.sparql.algebra.op.OpJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpLeftJoin;
import com.hp.hpl.jena.sparql.algebra.op.OpOrder;
import com.hp.hpl.jena.sparql.algebra.op.OpProject;
import com.hp.hpl.jena.sparql.algebra.op.OpReduced;
import com.hp.hpl.jena.sparql.algebra.op.OpService;
import com.hp.hpl.jena.sparql.algebra.op.OpSlice;
import com.hp.hpl.jena.sparql.algebra.op.OpUnion;
import com.hp.hpl.jena.sparql.core.Var;

public class WellDesignedService {

	/**
	 * Applies the well-designedness test described in Semantics and Complexity of SPARQL by Perez et al, 2009.
	 * 
	 * @param q
	 *            The query to be tested
	 * @return
	 */
	public static boolean isWellDesigned(Query q) {
		Op op = Algebra.compile(q);
		return isWellDesigned(op);
	}

	/**
	 * Applies the well-designedness test described in Semantics and Complexity of SPARQL by Perez et al, 2009.
	 * 
	 * @param op
	 *            The pattern to be tested
	 * @return True if op is well-designed
	 */
	public static boolean isWellDesigned(Op op) {
		return isWellDesignedWorker(op, op);
	}

	private static boolean isWellDesignedWorker(Op op, Op subop) {
		if (subop instanceof OpProject || subop instanceof OpSlice || subop instanceof OpReduced || subop instanceof OpOrder || subop instanceof OpDistinct || subop instanceof OpService) {
			Op sub = ((Op1) subop).getSubOp();
			return isWellDesignedWorker(sub, sub);
		}
		if (subop instanceof OpBGP) {
			return true;
		}
		if (subop instanceof OpUnion) {
			OpUnion union = (OpUnion) subop;
			return isWellDesignedWorker(union.getLeft(), union.getLeft()) && isWellDesignedWorker(union.getRight(), union.getRight());
		}
		if (subop instanceof OpJoin) {
			OpJoin join = (OpJoin) subop;
			return isWellDesignedWorker(op, join.getLeft()) && isWellDesignedWorker(op, join.getRight());
		}
		if (subop instanceof OpLeftJoin) {
			// for every subpattern P'=P1 opt P2:
			// calculate Vp1, Vp2 and Op, which is all vars outside P'
			OpLeftJoin lj = (OpLeftJoin) subop;
			Op right = lj.getRight();
			if (lj.getExprs() != null) {
				right = OpFilter.filterDirect(lj.getExprs(), lj.getRight());
			}
			Set<Var> Vp1 = new HashSet<Var>(OpVars.mentionedVars(lj.getLeft()));
			Set<Var> Vp2 = new HashSet<Var>(OpVars.mentionedVars(right));
			Set<Var> Op = new HashSet<Var>(SubsumptionService.getOtherVars(op, subop));

			// the property holds if Vp2 intersect Op is a subset of Vp1
			Vp2.retainAll(Op);
			return Vp1.containsAll(Vp2) && isWellDesignedWorker(op, lj.getLeft()) && isWellDesignedWorker(op, right);

		}
		if (subop instanceof OpFilter) {
			// for every subpattern P'=P Filter R, check that vars(R) is a
			// subset
			// of vars(P) including it's parent
			OpFilter f = (OpFilter) subop;
			Set<Var> filtervars = f.getExprs().getVarsMentioned();
			Set<Var> opvars = new HashSet<Var>(OpVars.mentionedVars(f.getSubOp()));
			return opvars.containsAll(filtervars) && isWellDesignedWorker(op, f.getSubOp());
		}

		throw new UnsupportedOperationException(subop.toString());

	}

}
