package spam.transformationRules;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import spam.model.TreePattern;
import spam.model.TreePatternNode;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.NodeFactory;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.sparql.algebra.Algebra;
import com.hp.hpl.jena.sparql.algebra.OpVars;
import com.hp.hpl.jena.sparql.algebra.op.OpBGP;
import com.hp.hpl.jena.sparql.core.BasicPattern;
import com.hp.hpl.jena.sparql.core.Var;
import com.hp.hpl.jena.sparql.engine.QueryIterator;
import com.hp.hpl.jena.sparql.engine.binding.Binding;
import com.hp.hpl.jena.sparql.expr.Expr;
import com.hp.hpl.jena.sparql.graph.GraphFactory;

public class TransformationUtilities {

	/**
	 * Deletion of redundant triples
	 *
	 * @param tree
	 *            the tree pattern
	 * @return true if there were any changes
	 */
	public static boolean R1(TreePattern tree) {
		return R1worker(tree.getRoot(), new HashSet<Triple>());
	}

	private static boolean R1worker(TreePatternNode node, Set<Triple> seen) {
		boolean changed = false;

		// first we remove triples that may be repeated within this node
		BasicPattern clean = new BasicPattern();
		for (Triple t : node.getBgp().getPattern().getList()) {
			if (!clean.getList().contains(t)) {
				clean.add(t);
			} else {
				changed = true;
			}
		}

		if (changed) {
			node.setBgp(new OpBGP(clean));
		}

		// then we remove triples that we've already seen
		changed |= node.getBgp().getPattern().getList().removeAll(seen);

		// check if there are any triples left
		if (node.getBgp().getPattern().isEmpty()) {
			node.delete();
		}

		// we need to create a new set to add triples; otherwise the recursion
		// goes crazy in parallel branches since triples are never removed from
		// the
		// set when the function returns
		Set<Triple> seen2 = new HashSet<Triple>(seen);
		seen2.addAll(node.getBgp().getPattern().getList());

		// since children can modify the parent's children list, we iterate over
		// a shallow copy of it
		ArrayList<TreePatternNode> childrenClone = new ArrayList<TreePatternNode>(node.getChildren());
		for (TreePatternNode child : childrenClone) {
			changed |= R1worker(child, seen2);
		}

		return changed;
	}

	/**
	 * check if R1 is applicable
	 *
	 * @param tree
	 * @return
	 */
	public static boolean R1check(TreePattern tree) {
		return R1checkWorker(tree.getRoot(), new HashSet<Triple>());
	}

	private static boolean R1checkWorker(TreePatternNode node, Set<Triple> seen) {
		// first we check for triples that may be repeated within this node
		BasicPattern clean = new BasicPattern();
		for (Triple t : node.getBgp().getPattern().getList()) {
			if (!clean.getList().contains(t)) {
				clean.add(t);
			} else {
				return true;
			}
		}

		// then we check if this node has a triple we've seen before
		for (Triple t : node.getBgp().getPattern().getList()) {
			if (seen.contains(t)) {
				return true;
			}
		}

		Set<Triple> seen2 = new HashSet<Triple>(seen);
		seen2.addAll(node.getBgp().getPattern().getList());

		// check for all children
		for (TreePatternNode child : node.getChildren()) {
			if (R1checkWorker(child, seen2)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Delete nodes which introduce no new variables
	 *
	 * @param tree
	 * @return true if the application of R2 produced any changes
	 */
	public static boolean R2(TreePattern tree) {
		return R2worker(tree.getRoot(), new HashSet<Var>());
	}

	private static boolean R2worker(TreePatternNode node, Set<Var> seen) {
		boolean changed = false;

		// check for new variables
		Set<Var> vars = new HashSet<Var>(OpVars.mentionedVars(node.getBgp()));
		if (!node.isRoot() && seen.containsAll(vars)) {
			node.mergeIntoChildren();
			changed = true;
		}

		// add new variables to the set
		Set<Var> seen2 = new HashSet<Var>(seen);
		seen2.addAll(vars);

		ArrayList<TreePatternNode> childrenClone = new ArrayList<TreePatternNode>(node.getChildren());
		for (TreePatternNode child : childrenClone) {
			changed |= R2worker(child, seen2);
		}
		return changed;
	}

	/**
	 * Applies R3 to the tree. Returns true if the tree was modified. This should be called in a while loop to ensure that the rule is applied exhaustively: while(R3(tree));
	 *
	 * @param tree
	 * @return
	 */
	public static boolean R3(TreePattern tree) {
		return R3worker(tree.getRoot(), new HashSet<Triple>(), new HashSet<Var>());
	}

	private static boolean R3worker(TreePatternNode node, Set<Triple> branch, Set<Var> branchVars) {
		Set<Var> intersection = new HashSet<Var>(OpVars.mentionedVars(node.getBgp()));
		intersection.retainAll(branchVars);

		if (checkHomomorphism(node.getBgp(), branch, intersection)) {
			node.mergeIntoParent();
			return true;
		}

		// recurse into children
		Set<Triple> branch2 = new HashSet<Triple>(branch);
		Set<Var> branchVars2 = new HashSet<Var>(branchVars);
		branch2.addAll(node.getBgp().getPattern().getList());
		branchVars2.addAll(OpVars.mentionedVars(node.getBgp()));

		for (TreePatternNode child : node.getChildren()) {
			if (R3worker(child, branch2, branchVars2)) {
				return true;
			}
		}

		return false;
	}

	/**
	 *
	 * @param tree
	 * @return
	 */
	public static boolean RFilter(TreePattern tree) {
		return RFilterWorker(tree.getRoot());
	}

	private static boolean RFilterWorker(TreePatternNode node) {
		boolean change = false;
		if (node.getFilter() == null) {
			return false;
		}
		for (TreePatternNode child : node.getChildren()) {

			expr: for (Expr exp : node.getFilter()) {
				for (Var x : exp.getVarsMentioned()) {
					if (!OpVars.mentionedVars(child.getBgp()).contains(x)) {
						continue expr;
					}
				}
				// if we got here, then the child mentions every variable in exp
				change = true;
				child.addFilter(exp);
			}

			change |= RFilterWorker(child);
		}
		return change;
	}

	/**
	 * Checks if there exists a homomorphism h from P1 to P2 such that for every ?X in identityVariables, h(?X)=?X.
	 *
	 * @param P1
	 * @param P2
	 * @param identityVariables
	 * @return
	 */
	public static boolean checkHomomorphism(OpBGP P1, Set<Triple> P2, Set<Var> identityVariables) {

		Graph g = GraphFactory.createDefaultGraph();
		for (Triple t : P2) {

			Node s = stringifyVar(t.getSubject()), p = stringifyVar(t.getPredicate()), o = stringifyVar(t.getObject());

			Triple tmod = new Triple(s, p, o);
			g.add(tmod);

		}

		QueryIterator r1 = Algebra.exec(P1, g);

		for (; r1.hasNext();) {
			r1.nextBinding();

		}
		r1.close();

		// check for homomorphisms
		QueryIterator res = Algebra.exec(P1, g);

		// now check that h(?x)=?x for some homomorphism in res
		bindings: while (res.hasNext()) {
			Binding b = res.nextBinding();
			for (Var v : OpVars.mentionedVars(P1)) {
				if (!b.contains(v)) {
					continue bindings;
				}
			}
			for (Var v : identityVariables) {
				if (!b.contains(v) || !b.get(v).toString().equals("\"" + v.toString() + "\"")) {
					continue bindings;
				}
			}
			// if we got here, then this binding makes h(?x)=?x for every ?x in
			// identityVariables
			res.close();
			return true;
		}
		return false;
	}

	/**
	 * Transforms nodes into literals if they are variables.
	 * 
	 * @param n
	 * @return
	 */
	private static Node stringifyVar(Node n) {
		return n.isVariable() ? NodeFactory.createLiteral(n.toString()) : n;
	}

	public static List<RuleApplication> getApplicableRules(TreePattern tree) {
		List<RuleApplication> rules = new ArrayList<RuleApplication>();

		rules.addAll(R1Application.findAll(tree));
		rules.addAll(R2Application.findAll(tree));
		rules.addAll(R3Application.findAll(tree));
		rules.addAll(R4Application.findAll(tree));
		return rules;
	}
}
